import os
import numpy as np

# from mlc.dataloaders.generative.ddpm import DDPMDataLoader

from mlc.datasets.dataset_factory import load_dataset
from mlc.metrics.compute import compute_metric, compute_metrics
from mlc.metrics.metric_factory import create_metric
from mlc.models.model_factory import load_model
from mlc.transformers.tab_scaler import TabScaler
import torch


models_params = {
    "tabtransformer": {
        "model_name": "TabTransformer",
        "dim": 8,
        "batch_size": 64,
        # "num_classes": 1,
        "depth": 6,
        "heads": 4,
        "weight_decay": -2,
        "learning_rate": -3,
        "dropout": 0.1,
        "data_parallel": False,
    },
    "deepfm": {
        "dnn_dropout": 0.3,
        "num_dense_features": None,
    },
    "torchrln": {
        "n_layers": 5,
        "theta": -12,
        "learning_rate": 0.0005588211696108453,
        "norm": 2,
        "hidden_dim": 78,
    },
}

other = {'depth': 3, 'dim': 32, 'dropout': 0.1, 'heads': 2, 'learning_rate': -3, 'weight_decay': -6}
# other = {'depth': 12, 'dim': 256, 'dropout': 0.3, 'heads': 4, 'learning_rate': -4, 'weight_decay': -2}

def run(
    dataset_name: str = "lcld_201317_ds_time",
    # dataset_name: str = "url",
    model_name: str = "tabtransformer",
) -> None:

    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    common_model_params = {
        "x_metadata": metadata,
        "objective": "binary",
        "use_gpu": True,
        "batch_size": 64,
        "num_classes": 2,
        "epochs": 1,
        "early_stopping_rounds": 1,
        "val_batch_size": 1024,
        "class_weight": "balanced",
    }

    x, y = dataset.get_x_y()
    splits = dataset.get_splits()

    x_train = x.iloc[splits["train"]]
    y_train = y[splits["train"]]
    x_test = x.iloc[splits["test"]]
    y_test = y[splits["test"]]
    x_val = x.iloc[splits["val"]]
    y_val = y[splits["val"]]

    x_train = x_train.iloc[:200000]
    y_train = y_train[:200000]
    x_test = x_test.iloc[:50000]
    y_test = y_test[:50000]
    x_val = x_val[:50000]
    y_val = y_val[:50000]
    
    # x_train = x.iloc[:350000]
    # y_train = y[:350000]
    # x_test = x.iloc[400000:]
    # y_test = y[400000:]
    # x_val = x.iloc[350000:400000]
    # y_val = y[350000:400000]

    # {'depth': 12, 'dim': 32, 'dropout': 0.5, 'heads': 2, 'learning_rate': -3, 'weight_decay': -3}
    model_class = load_model(model_name)
    model_params = {**models_params[model_name], **common_model_params}
    
    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )

    # model_params = {**models_params[model_name], **common_model_params, **other}
    model = model_class(**model_params, scaler = scaler)
    # scaler = TabScaler(num_scaler="min_max", one_hot_encode=False)
    # scaler.fit(x.values, x_type=metadata["type"])

    # model.set_dataloader(
    #     DDPMDataLoader(original_dataset=dataset)
    # )
    metadata = dataset.get_metadata(only_x=True)
    # {'depth': 12, 'dim': 32, 'dropout': 0.5, 'heads': 2, 'learning_rate': -3, 'weight_decay': -3}
    # model_class = load_model("tabtransformer")
    # model = model_class(
    #     x_metadata=metadata,
    #     objective="binary",
    #     use_gpu=True,
    #     dim=8,
    #     batch_size=64,
    #     num_classes=1,
    #     depth=6,
    #     heads=4,
    #     weight_decay=-2,
    #     learning_rate=-3,
    #     dropout=0.3,
    #     data_parallel=False,
    #     val_batch_size=64,
    #     epochs=5,
    #     model_name="TabTransformer",
    #     # dataset="lcld_v2_iid",
    #     early_stopping_rounds=1,
    # )

   
    # model_class = load_model("deepfm")

    # model = model_class(
    #     x_metadata=metadata,
    #     objective="binary",
    #     use_gpu=False,
    #     batch_size=64,
    #     epochs=1,
    #     early_stopping_rounds=1,
    #     dnn_dropout=0.3,
    #     num_dense_features=None,
    # )

    # model_class = load_model("torchrln")

    # model = model_class(
    #     x_metadata=metadata,
    #     objective="binary",
    #     use_gpu=False,
    #     batch_size=64,
    #     epochs=10,
    #     early_stopping_rounds=2,
    #     num_classes=2,
    #     n_layers=5,
    #     theta=-12,
    #     learning_rate=0.0005588211696108453,
    #     class_weight="balanced",
    #     norm=2,
    #     hidden_dim=78,
    #     scaler=scaler,
    # )

    # model_class = load_model("saint")
    # model = model_class(
    #     objective="binary",
    #     x_metadata=metadata,
    #     batch_size=64,
    #     epochs=1,
    #     early_stopping_rounds=1,
    #     num_classes=1,
    # )

    # model_class = load_model("vime")
    # model = model_class(
    #     objective="binary",
    #     x_metadata=metadata,
    #     batch_size=64,
    #     epochs=1,
    #     early_stopping_rounds=1,
    #     num_classes=1,
    #     device="cuda",
    # )

    # model.set_dataloader(
    #     SemiSupervisedDataLoader(take_amount=1000)
    # )

    # model.fit(
    #     x_train.values,
    #     y_train,
    #     x_val.values,
    #     y_val,
    # )

    # model.fit(
    #     scaler.transform(x_train.values),
    #     # y_train,
    #     np.array([1 - y_train, y_train]).T,
    #     scaler.transform(x_val.values),
    #     # y_val
    #     np.array([1 - y_val, y_val]).T,
    # )
    save_path = "./data/models/lcld_tabtransformer.model"

    model.fit(
        x_train.values,
        # scaler.transform(x_train.values),
        # y_train,
        np.array([1 - y_train, y_train]).T,
        x_val.values,
        # scaler.transform(x_val.values),
        # y_val
        np.array([1 - y_val, y_val]).T,
    )

    model.save(save_path)
    model = model_class.load_class(
        save_path, x_metadata=metadata, scaler=scaler
    )

    metrics = ["auc", "accuracy", "mcc"]
    metrics_obj = [create_metric(m) for m in metrics]
    res = compute_metrics(model, metrics_obj, x_test.values, y_test)

    for i, e in enumerate(metrics):
        print(f"Test {e}: {res[i]}")


if __name__ == "__main__":
    run()



import sys

sys.path.append("../constrained-attacks")

from typing import Tuple
import numpy as np
import torch
from torch.utils.data import DataLoader, TensorDataset

from mlc.models.torch_models import BaseModelTorch
from mlc.dataloaders.dataloader import BaseDataLoader
from constrained_attacks.attacks.cta.capgd import CAPGD
from mlc.datasets.dataset import Dataset

attacks = {
    "pgd": CAPGD, "c-pgd": CAPGD
}


def get_adversarial_collate(attack, augment, custom_loader):
    def custom_collate(batch):
        collated = torch.utils.data.dataloader.default_collate(batch)
        if attack is not None:
            x = collated[0].to(attack.device)
            y = collated[1].to(attack.device)
            y_adv = y if custom_loader.dataloader.transform_y is None else y[:, 1].long()
            adv_x = attack(x,
                           y_adv) if custom_loader.dataloader.scaler is None else custom_loader.dataloader.scaler.transform(
                attack(custom_loader.scaler.inverse_transform(x), y_adv))
        if augment == 2:
            return torch.cat([x, adv_x]), *[torch.cat([collated[k], collated[k]]) for k in range(1, len(collated))]
        elif augment == 1:
            return torch.cat([x[:len(x) // 2], adv_x[len(x) // 2:]]), *collated[1:]
        else:
            return adv_x, *collated[1:]

    return custom_collate


class MadryATDataLoader(BaseDataLoader):
    def __init__(self, dataset: Dataset = None, scaler=None, model: BaseModelTorch = None, attack="pgd",
                 attack_params={"eps": 0.3, "norm": "L2"}, move_all_to_model_device: bool = True,
                 filter_constraints: bool = False, augment: int = 1, verbose=False
                 ) -> None:
        self.num_workers = 0
        self.dataset = dataset
        self.device = model.device
        self.model = model.wrapper_model
        self.move_all_to_model_device = move_all_to_model_device
        self.custom_scaler = None

        constraints = dataset.get_constraints()
        fix_equality_constraints_end = True
        if not attack.startswith("c-"):
            constraints.relation_constraints = None
            fix_equality_constraints_end = False

        self.attack = attacks.get(attack, CAPGD)(constraints=constraints, scaler=scaler, model=self.model,
                                                 fix_equality_constraints_end=fix_equality_constraints_end,
                                                 fix_equality_constraints_iter=False,
                                                 model_objective=model.predict_proba, **attack_params,
                                                 verbose=verbose)
        self.filter_constraints = filter_constraints
        self.augment = augment
        self.scaler = scaler

    def set_tensors(self, tensors, scaler):
        self.dataset.tensors = tensors
        self.custom_scaler = scaler

    def get_dataloader(
            self, x: np.ndarray, y: np.ndarray, train: bool, batch_size: int) -> DataLoader[Tuple[torch.Tensor, ...]]:
        # Create dataset
        x_scaled = torch.Tensor(self.scaler.transform(x)).float()
        x = torch.Tensor(x).float()
        y = torch.Tensor(y).long()
        if self.move_all_to_model_device:
            x = x.to(self.device)
            y = y.to(self.device)

        self.dataset = TensorDataset(x, y)

        # Create dataloader
        if train:

            self.dataloader = DataLoader(
                dataset=self.dataset,
                batch_size=batch_size,
                shuffle=True,
                num_workers=self.num_workers,
                collate_fn=get_adversarial_collate(self.attack, self.augment, self)
            )

        else:
            self.dataloader = DataLoader(
                dataset=self.dataset,
                batch_size=batch_size,
                shuffle=False,
                num_workers=self.num_workers
            )
        self.dataloader.scaler = None
        self.dataloader.transform_y = None
        return self.dataloader



import os
import sys

sys.path.append(".")
from mlc.logging import XP

import torch

from argparse import ArgumentParser, Namespace
from mlc.datasets.dataset_factory import load_dataset

from mlc.dataloaders.generative.ctgan import CTGANDataLoader
# from mlc.dataloaders.generative.ddpm import DDPMDataLoader
from mlc.transformers.tab_scaler import TabScaler

AUGMENTATIONS = {"ctgan": CTGANDataLoader}  # , "ddpm": DDPMDataLoader}


def run(augmentation_name, dataset_name, model_name, train_batch_size=32,train_generator_epochs=50, debug=0):
    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    x, y = dataset.get_x_y()

    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )
    filter_constraints = False
    path = "../models/mlc/data_augmentation"
    save_path = os.path.join(path, f"{dataset_name}_{augmentation_name}.pt")
    pretrained_path = save_path if os.path.exists(save_path) else None
    os.makedirs(path, exist_ok=True)

    experiment = XP(dict(dataset=dataset_name, model_name=model_name), project_name="DataAugment")
    augmentation = AUGMENTATIONS[augmentation_name](original_dataset=dataset, scaler=scaler,
                                                    pretrained_path=pretrained_path, save_path=save_path,
                                                    filter_constraints=filter_constraints,
                                                    train_batch_size=train_batch_size,
                                                    generator_epochs=train_generator_epochs,
                                                    verbose=debug, subset=10000 if debug else 0)

    augmented_loader = augmentation.get_dataloader(x, y, train=True, batch_size=train_batch_size, augment=False)
    experiment.log_parameters({"size_trainining": len(augmented_loader)})

    for batch in augmented_loader:
        print(batch)


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Training with Hyper-parameter optimization"
    )
    parser.add_argument("--dataset_name", type=str, default="lcld_v2_iid",
                        )
    parser.add_argument("--augmentation_name", type=str, default="ctgan",
                        )
    parser.add_argument("--model_name", type=str, default="tabtransformer",
                        )

    parser.add_argument("--train_batch_size", type=int, default=1024)

    parser.add_argument("--train_generator_epochs", type=int, default=500)

    parser.add_argument("--debug", type=int, default=0)

    args = parser.parse_args()

    run(augmentation_name=args.augmentation_name, dataset_name=args.dataset_name, model_name=args.model_name,
        train_batch_size=args.train_batch_size,train_generator_epochs=args.train_generator_epochs, debug=args.debug)



from __future__ import annotations

import abc
from typing import Any, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.base import TransformerMixin


class Transformer(TransformerMixin):
    def __init__(self, **kwargs: Any) -> None:
        self.name = kwargs["name"]
        self.constructor_kwargs = kwargs

    @abc.abstractmethod
    def fit(
        self,
        x: Union[npt.NDArray[np.float_], pd.DataFrame],
        y: Union[npt.NDArray[np.int_], pd.Series] = None,
    ) -> Transformer:
        pass

    @abc.abstractmethod
    def transform(
        self, x: Union[npt.NDArray[np.float_], pd.DataFrame]
    ) -> Union[npt.NDArray[np.float_], pd.DataFrame]:
        pass

    def clone(self) -> Transformer:
        return self.__class__(**self.constructor_kwargs)

    @abc.abstractmethod
    def load(self, path: str) -> None:
        pass

    @abc.abstractmethod
    def save(self, path: str) -> None:
        pass



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import torch
import torch.nn as nn
from sklearn.preprocessing import OneHotEncoder

from mlc.typing import NDInt, NDNumber
from mlc.utils import round_func_BPDA, to_numpy_number, to_torch_number
import joblib


def get_num_idx(x_type: Union[pd.Series, npt.NDArray[np.str_]]) -> List[int]:
    return np.where(x_type != "cat")[0].tolist()


def get_cat_idx(x_type: Union[pd.Series, npt.NDArray[np.str_]]) -> List[int]:
    return np.where(x_type == "cat")[0].tolist()


def copy_x(
    x: Union[npt.NDArray[Any], torch.Tensor]
) -> Union[npt.NDArray[Any], torch.Tensor]:
    if isinstance(x, np.ndarray):
        return x.copy()
    if isinstance(x, torch.Tensor):
        return torch.clone(x)
    raise NotImplementedError


def softargmax(x: torch.Tensor, dim: int = -1) -> torch.Tensor:
    # crude: assumes max value is unique
    # Can cause rounding errors
    beta = 100.0
    xx = beta * x
    sm = torch.nn.functional.softmax(xx, dim=dim)
    indices = torch.arange(x.shape[dim]).to(x.device)
    y = torch.mul(indices, sm)
    result = torch.sum(y, dim)
    return result


def _handle_zeros_in_scale_torch(
    scale: torch.Tensor,
    copy: bool = True,
    constant_mask: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """Set scales of near constant features to 1.
    The goal is to avoid division by very small or zero values.
    Near constant features are detected automatically by identifying
    scales close to machine precision unless they are precomputed by
    the caller and passed with the `constant_mask` kwarg.
    Typically for standard scaling, the scales are the standard
    deviation while near constant features are better detected on the
    computed variances which are closer to machine precision by
    construction.
    """
    # if we are fitting on 1D arrays, scale might be a scalar
    if scale.dim() == 0:
        if scale == 0.0:
            scale = torch.tensor(1.0)
        return scale
    else:
        if constant_mask is None:
            # Detect near constant values to avoid dividing by a very small
            # value that could lead to surprising results and numerical
            # stability issues.
            constant_mask = scale < 10 * torch.finfo(scale.dtype).eps

        if copy:
            # New array to avoid side-effects
            scale = scale.clone()
        scale[constant_mask] = 1.0
        return scale


def process_cat_idx_params(
    x: torch.Tensor,
    cat_idx: Optional[List[int]] = None,
    x_type: Optional[Union[pd.Series, npt.NDArray[np.str_]]] = None,
) -> Tuple[List[int], List[int]]:
    if cat_idx is None:
        if x_type is not None:
            cat_idx = get_cat_idx(x_type)

    nb_features = x.shape[1]
    num_idx = [*range(nb_features)]

    if cat_idx is None:
        cat_idx = []
    else:
        num_idx = [e for e in num_idx if e not in cat_idx]

    return num_idx, cat_idx


@dataclass
class ScalerData:
    x_min: Optional[torch.Tensor] = None
    x_max: Optional[torch.Tensor] = None
    x_mean: Optional[torch.Tensor] = None
    x_std: Optional[torch.Tensor] = None
    categories: List[NDInt] = field(default_factory=list)
    cat_idx: List[int] = field(default_factory=list)
    num_idx: List[int] = field(default_factory=list)


class TabScaler:
    def __init__(
        self,
        num_scaler: str = "min_max",
        one_hot_encode: bool = True,
        out_min: float = 0.0,
        out_max: float = 1.0,
    ) -> None:

        # Params
        self.num_scaler = num_scaler
        self.one_hot_encode = one_hot_encode
        self.out_min = out_min
        self.out_max = out_max

        # Internal
        self.fitted = False
        self.x_min: Optional[torch.Tensor] = None
        self.x_max: Optional[torch.Tensor] = None
        self.mean: Optional[torch.Tensor] = None
        self.std: Optional[torch.Tensor] = None
        self.categories: List[NDInt] = []
        self.cat_idx: List[int] = []
        self.num_idx: List[int] = []

        # Check params
        if self.out_min >= self.out_max:
            raise ValueError("out_min must be smaller than out_max")

    def do_one_hot_encode(self) -> bool:
        return self.one_hot_encode and len(self.cat_idx) > 0

    def fit(
        self,
        x: Union[torch.Tensor, NDNumber],
        cat_idx: Optional[List[int]] = None,
        x_type: Union[pd.Series, npt.NDArray[np.str_]] = None,
    ) -> TabScaler:

        if isinstance(x, pd.DataFrame):
            return self.fit(to_torch_number(x.values), cat_idx, x_type)

        if isinstance(x, np.ndarray):
            return self.fit(to_torch_number(x), cat_idx, x_type)

        # Process feature types
        self.num_idx, self.cat_idx = process_cat_idx_params(x, cat_idx, x_type)

        # Numerical features
        if self.num_scaler == "min_max":
            self.x_min = torch.min(x[:, self.num_idx], dim=0)[0]
            self.x_max = torch.max(x[:, self.num_idx], dim=0)[0]
            if self.x_min is None or self.x_max is None:
                raise ValueError("No numerical features to scale")
            self.min_max_scale = _handle_zeros_in_scale_torch(
                self.x_max - self.x_min
            )

        elif self.num_scaler == "standard":
            self.mean = torch.mean(x[:, self.num_idx], dim=0)
            self.std = _handle_zeros_in_scale_torch(
                torch.std(x[:, self.num_idx], dim=0)
            )

        elif self.num_scaler == "none":
            pass

        else:
            raise NotImplementedError

        # Categorical features
        if self.do_one_hot_encode():
            ohe = OneHotEncoder(sparse=False)
            ohe.fit(x.numpy()[:, self.cat_idx])
            self.categories = ohe.categories_

        self.fitted = True

        return self

    def fit_metadata(self, metadata: pd.DataFrame) -> TabScaler:

        # Process feature types
        self.num_idx, self.cat_idx = process_cat_idx_params(
            torch.tensor([metadata["min"]]), None, metadata["type"]
        )

        if self.num_scaler == "min_max":
            self.x_min = torch.tensor(
                metadata["min"].values.astype(np.float_)[self.num_idx],
                dtype=torch.float,
            )
            self.x_max = torch.tensor(
                metadata["max"].values.astype(np.float_)[self.num_idx],
                dtype=torch.float,
            )
            if self.x_min is None or self.x_max is None:
                raise ValueError("No numerical features to scale")
            self.min_max_scale = _handle_zeros_in_scale_torch(
                self.x_max - self.x_min
            )

        elif self.num_scaler == "standard":
            self.mean = torch.tensor(
                metadata["mean"].values.astype(np.float_)[self.num_idx],
                dtype=torch.float,
            )
            self.std = torch.tensor(
                metadata["std"].values.astype(np.float_)[self.num_idx],
                dtype=torch.float,
            )

        elif self.num_scaler == "none":
            pass

        else:
            raise NotImplementedError

        # Categorical features
        if self.do_one_hot_encode():
            self.categories = [
                np.arange(
                    int(metadata["min"].values[i]),
                    int(metadata["max"].values[i]) + 1,
                )
                for i in self.cat_idx
            ]

        self.fitted = True

        return self

    def fit_scaler_data(self, scaler_data: ScalerData) -> TabScaler:
        self.num_idx = scaler_data.num_idx
        self.cat_idx = scaler_data.cat_idx
        self.x_min = scaler_data.x_min
        self.x_max = scaler_data.x_max
        self.mean = scaler_data.x_mean
        self.std = scaler_data.x_std
        self.categories = scaler_data.categories
        self.fitted = True

        if self.x_min is not None and self.x_max is not None:
            self.min_max_scale = _handle_zeros_in_scale_torch(
                self.x_max - self.x_min
            )

        return self

    def get_scaler_data(self) -> ScalerData:
        return ScalerData(
            num_idx=self.num_idx,
            cat_idx=self.cat_idx,
            x_min=self.x_min,
            x_max=self.x_max,
            x_mean=self.mean,
            x_std=self.std,
            categories=self.categories,
        )

    def transform(
        self, x_in: Union[torch.Tensor, NDNumber]
    ) -> Union[torch.Tensor, NDNumber]:

        if not self.fitted:
            raise ValueError("Must fit scaler before transforming data")

        if isinstance(x_in, pd.DataFrame):
            return to_numpy_number(
                self.transform(to_torch_number(x_in.values))
            )

        if isinstance(x_in, np.ndarray):
            return to_numpy_number(self.transform(to_torch_number(x_in)))
        x = x_in.clone()

        # Numerical features

        if self.num_scaler == "min_max":
            x[:, self.num_idx] = (
                (x[:, self.num_idx].clone() - self.x_min.to(x.device))
                / self.min_max_scale.to(x.device)
            ) * (self.out_max - self.out_min) + self.out_min

        elif self.num_scaler == "standard":
            x[:, self.num_idx] = (
                x[:, self.num_idx] - self.mean.to(x.device)
            ) / self.std.to(x.device)

        elif self.num_scaler == "none":
            pass

        else:
            raise NotImplementedError

        if self.do_one_hot_encode():

            list_encoded = []

            for i, idx in enumerate(self.cat_idx):

                categories = self.categories[i]
                num_categories = len(categories)

                coder = torch.tensor(categories, dtype=torch.float32)
                codes = x[:, idx].clone()

                rows = x.shape[0]

                dummies = torch.broadcast_to(coder, (rows, num_categories))
                coded = codes.repeat_interleave(num_categories).reshape(
                    rows, num_categories
                )

                diff = coded - dummies.to(coded.device)
                diff[diff > 0] = -diff[diff > 0]
                elu = torch.nn.ELU()
                array_bef = elu(diff) + 0.55
                encoded = round_func_BPDA(array_bef)
                list_encoded.append(encoded)

            x = torch.concat([x[:, self.num_idx]] + list_encoded, dim=1)

        return x

    def inverse_transform(
        self, x_in: Union[torch.Tensor, NDNumber]
    ) -> Union[torch.Tensor, NDNumber]:

        if not self.fitted:
            raise ValueError("Must fit scaler before transforming data")

        device = "cpu"
        if isinstance(x_in, np.ndarray):
            return to_numpy_number(
                self.inverse_transform(to_torch_number(x_in))
            )
        else:
            device = x_in.device

        x = x_in.clone()

        # Categorical features
        if self.do_one_hot_encode():

            start_idx = len(self.num_idx)

            decoded_list = []
            for i, categories in enumerate(self.categories):
                num_categories = len(categories)
                end_idx = start_idx + num_categories

                decoded = softargmax(x[:, start_idx:end_idx], 1)
                decoded_list.append(decoded)
                start_idx = end_idx

            out: List[torch.Tensor] = []

            # First numerical features
            out.append(x[:, : self.cat_idx[0]])
            last_num_used = self.cat_idx[0]

            for i, idx in enumerate(self.cat_idx):
                out.append(decoded_list[i].reshape(-1, 1))
                if i < len(self.cat_idx) - 1:
                    num_to_add = self.cat_idx[i + 1] - idx - 1
                else:
                    num_to_add = len(self.num_idx) - last_num_used
                end_idx = last_num_used + num_to_add
                out.append(x[:, last_num_used:end_idx])
                last_num_used = end_idx

            x = torch.concat(out, dim=1)

        # Numerical features
        if self.num_scaler == "min_max":
            x[:, self.num_idx] = (x[:, self.num_idx] - self.out_min) / (
                self.out_max - self.out_min
            )
            x[:, self.num_idx] = x[:, self.num_idx] * self.min_max_scale.to(
                x.device
            ) + self.x_min.to(x.device)

        elif self.num_scaler == "standard":
            x[:, self.num_idx] = x[:, self.num_idx] * self.std + self.mean

        elif self.num_scaler == "none":
            pass

        else:
            raise NotImplementedError

        return x if isinstance(x_in, np.ndarray) else x.to(device)

    def get_transorm_nn(self) -> nn.Module:
        return Transform(self)

    def transform_mask(self, mask: torch.Tensor) -> torch.Tensor:
        mask_out = [mask[self.num_idx]]
        for i, e in enumerate(self.cat_idx):
            mask_out.append(mask[e].repeat(len(self.categories[i])))
        return torch.cat(mask_out, dim=0)

    def get_transformed_num_features(
        self,
    ) -> int:
        return len(self.num_idx) + sum([len(e) for e in self.categories])

    def save(self, path: str) -> None:
        joblib.dump(self.get_scaler_data(), path)

    def load(self, path: str) -> None:
        self.fit_scaler_data(joblib.load(path))


class Transform(nn.Module):
    def __init__(self, scaler: TabScaler):
        self.scaler = scaler
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return to_torch_number(self.scaler.transform(x))



from abc import ABC

import joblib

from mlc.models.model import Model


class SkModel(Model, ABC):
    def load(self, path: str) -> None:
        n_jobs = self.model.get_params().get("n_jobs")
        self.model = joblib.load(path)
        self.model.set_params(n_jobs=n_jobs)

    def save(self, path: str) -> None:
        joblib.dump(self.model, path, compress=3)



from typing import Any, Dict

import joblib
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor as SKRandomForestRegressor
from sklearn.pipeline import Pipeline

from mlc.datasets.processing import get_numeric_categorical_preprocessor
from mlc.models.model import Model


class RandomForest(Model):
    def __init__(
        self,
        x_metadata: pd.DataFrame,
        name: str = "rf",
        **kwargs: Dict[str, Any],
    ) -> None:
        self.x_metadata = x_metadata
        self.rf_args = kwargs
        cat_features = self.x_metadata["type"] == "cat"
        num_features = ~cat_features
        self.model = Pipeline(
            steps=[
                (
                    "pre_process",
                    get_numeric_categorical_preprocessor(
                        numeric_features=self.x_metadata[num_features][
                            "feature"
                        ],
                        categorical_features=self.x_metadata[cat_features][
                            "feature"
                        ],
                    ),
                ),
                ("model", RandomForestClassifier(**self.rf_args)),
            ]
        )
        super(RandomForest, self).__init__(name)

    def fit(self, x, y):
        self.model.fit(x, y)

    def predict(self, x):
        return self.model.predict_proba(x)

    def load(self, path):
        self.model = joblib.load(path)

    def save(self, path):
        joblib.dump(self.model, path)


class RandomForestBalanced(Model):
    def __init__(self, x_metadata, **kwargs):
        name = "rf_balanced"
        self.model = RandomForest(
            x_metadata=x_metadata, class_weight="balanced", **kwargs
        )
        super(RandomForestBalanced, self).__init__(name)

    def fit(self, x, y):
        self.model.fit(x, y)

    def predict(self, x):
        return self.model.predict(x)

    def load(self, path):
        self.model.load(path)

    def save(self, path):
        self.model.save(path)


class RandomForestRegressor(Model):
    def __init__(self, x_metadata, name="rf_regressor", **kwargs):

        self.x_metadata = x_metadata
        self.rf_args = kwargs
        cat_features = self.x_metadata[self.x_metadata["type"] == "cat"]
        num_features = ~cat_features
        self.model = Pipeline(
            steps=[
                (
                    "pre_process",
                    get_numeric_categorical_preprocessor(
                        numeric_features=self.x_metadata[num_features][
                            "feature"
                        ],
                        categorical_features=self.x_metadata[cat_features][
                            "feature"
                        ],
                    ),
                ),
                ("model", SKRandomForestRegressor(**self.rf_args)),
            ]
        )
        super(RandomForestRegressor, self).__init__(name)

    def fit(self, x, y):
        self.model.fit(x, y)

    def predict(self, x):
        return self.model.predict(x)

    def load(self, path):
        self.model = joblib.load(path)

    def save(self, path):
        joblib.dump(self.model, path)


models = [
    ("rf", RandomForest),
    ("rf_balanced", RandomForestBalanced),
    ("rf_regressor", RandomForestRegressor),
]



from typing import Dict, List

import numpy as np
import numpy.typing as npt
from sklearn.model_selection import train_test_split

from mlc.constraints.relation_constraint import BaseRelationConstraint, Feature
from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DataSource,
    DefaultIndexSorter,
    DownloadFileDataSource,
    Splitter,
    Task,
)


class WidsSplitter(Splitter):
    def get_splits(self, dataset: Dataset) -> Dict[str, npt.NDArray[np.int_]]:
        _, y = dataset.get_x_y()
        i = np.arange(len(y))

        i_train, i_test = train_test_split(
            i,
            random_state=100,
            shuffle=True,
            stratify=y[i],
            test_size=0.2,
        )
        i_train, i_val = train_test_split(
            i_train,
            random_state=200,
            shuffle=True,
            stratify=y[i_train],
            test_size=0.2,
        )
        return {"train": i_train, "val": i_val, "test": i_test}


def get_relation_constraints(
    metadata: DataSource,
) -> List[BaseRelationConstraint]:

    # features = metadata.load_data()["feature"].to_list()

    # Maybe update the index with feature names in the future
    g_min_max = []
    for i in range(33, 94, 2):
        g_min_max.append(Feature(i + 1) <= Feature(i))
    return g_min_max


def create_dataset() -> Dataset:
    data_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "ESKij8CqW6dAlIe3bW-6rk0BCyDZ7B1gLYI3OTLdlH2wGg?download=1",
        file_data_source=CsvDataSource(path="./data/mlc/wids/wids.csv"),
    )
    metadata_source = DownloadFileDataSource(
        url="https://uniluxembourg-my.sharepoint.com/:x:/g/personal/"
        "thibault_simonetto_uni_lu/"
        "EWZgana5BVJNgvCiM1I-Y7MBhXOw7MrxyjGWMg2FRnMCBA?download=1",
        file_data_source=CsvDataSource(
            path="./data/mlc/wids/wids_metadata.csv"
        ),
    )
    tasks = [
        Task(
            name="hospital_death",
            task_type="classification",
            evaluation_metric="mcc",
        )
    ]
    sorter = DefaultIndexSorter()
    splitter = WidsSplitter()
    relation_constraints = get_relation_constraints(metadata_source)

    wids = Dataset(
        name="wids",
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=tasks,
        sorter=sorter,
        splitter=splitter,
        relation_constraints=relation_constraints,
    )
    return wids


datasets = [
    {
        "name": "wids",
        "fun_create": create_dataset,
    }
]



from typing import List, Union

import numpy as np
import numpy.typing as npt
from sklearn.preprocessing import OneHotEncoder

from mlc.metrics.metric import Metric
from mlc.metrics.metrics import PredClassificationMetric
from mlc.models.model import Model
from mlc.typing import NDInt, NDNumber


def default_model_prediction(
    model: Model, x: npt.NDArray[np.float_]
) -> Union[npt.NDArray[np.int_], npt.NDArray[np.float_]]:
    if model.objective in ["regression"]:
        return model.predict(x)
    if model.objective in ["binary", "classification"]:
        return model.predict_proba(x)
    raise NotImplementedError


def compute_metric(
    model: Model,
    metric: Metric,
    x: npt.NDArray[np.float_],
    y: npt.NDArray[Union[np.float_, np.int_]],
) -> npt.NDArray[np.generic]:
    return compute_metrics(model, metric, x, y)


def compute_metrics(
    model: Model,
    metrics: Union[Metric, List[Metric]],
    x: NDNumber,
    y: NDInt,
) -> npt.NDArray[np.generic]:

    if isinstance(metrics, Metric):
        return compute_metrics(model, [metrics], x, y)[0]

    y_score = default_model_prediction(model, x)

    return compute_metrics_from_scores(metrics, y, y_score)


def compute_metrics_from_scores(
    metrics: Union[Metric, List[Metric]],
    y: npt.NDArray[Union[np.float_, np.int_]],
    y_score: npt.NDArray[np.float_],
) -> npt.NDArray[np.generic]:

    if isinstance(metrics, Metric):
        return compute_metrics_from_scores([metrics], y, y_score)[0]

    y_2d = OneHotEncoder(sparse=False).fit_transform(y[:, None])

    y_pred = np.argmax(y_score, axis=1)
    out = []
    for metric in metrics:
        if isinstance(metric, PredClassificationMetric):
            out.append(metric.compute(y, y_pred))
        else:
            out.append(metric.compute(y_2d, y_score))

    return np.array(out)



from typing import Callable, Dict, List, Union

from mlc.datasets.dataset import Dataset
from mlc.datasets.samples.airlines import datasets as airlines_datasets
from mlc.datasets.samples.ctu_13_neris import datasets as ctu_13_neris_datasets
from mlc.datasets.samples.electricity import datasets as electricity_datasets
from mlc.datasets.samples.lcld import datasets as lcld_datasets
from mlc.datasets.samples.malware import datasets as malware_datasets
from mlc.datasets.samples.url import datasets as url_datasets
from mlc.datasets.samples.wids import datasets as wids_datasets

datasets: List[Dict[str, Union[str, Callable[[], Dataset]]]] = (
    lcld_datasets
    + ctu_13_neris_datasets
    + electricity_datasets
    + airlines_datasets
    + url_datasets
    + malware_datasets
    + wids_datasets
)


def load_dataset(dataset_name: str) -> Dataset:

    if dataset_name not in [e["name"] for e in datasets]:
        raise NotImplementedError("Dataset not available.")

    return [e["fun_create"]() for e in datasets if e["name"] == dataset_name][
        0
    ]



from abc import abstractmethod

import numpy as np
import numpy.typing as npt


class Metric:
    @abstractmethod
    def compute(
        self, y_true: npt.NDArray[np.generic], y_score: npt.NDArray[np.generic]
    ) -> npt.NDArray[np.generic]:
        pass



from __future__ import annotations

import abc
from abc import ABC
from typing import Any, Optional, Union

import numpy as np
import numpy.typing as npt

from mlc.typing import NDFloat, NDNumber
from torch.utils.data import DataLoader

class Model(ABC):
    def __init__(self, **kwargs: Any) -> None:
        self.name = kwargs["name"]
        self.objective = kwargs["objective"]
        self.constructor_kwargs = kwargs
        self.model: Optional[Any] = None

    def predict(
        self, x: npt.NDArray[np.float_]
    ) -> Union[npt.NDArray[np.int_], npt.NDArray[np.float_]]:
        """
        Returns the regression value or the concrete classes of binary /
        multi-class-classification tasks.
        (Save predictions to self.predictions)

        :param x: test data
        :return: predicted values / classes of test data (Shape N x 1)
        """

        if self.constructor_kwargs["objective"] == "regression":
            predictions = self.model.predict(x)
        elif (
            self.constructor_kwargs["objective"] == "classification"
            or self.constructor_kwargs["objective"] == "binary"
        ):
            prediction_probabilities = self.predict_proba(x)
            predictions = np.argmax(prediction_probabilities, axis=1)
        else:
            raise NotImplementedError

        return predictions

    def predict_proba(
        self, x: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        """
        Only implemented for binary / multi-class-classification tasks.
        Returns the probability distribution over the classes C.
        (Save probabilities to self.prediction_probabilities)

        :param x: test data
        :return: probabilities for the classes (Shape N x C)
        """

        prediction_probabilities = self.model.predict_proba(x)

        # If binary task returns only probability for the true class,
        # adapt it to return (N x 2)
        if prediction_probabilities.shape[1] == 1:
            prediction_probabilities = np.concatenate(
                (
                    1 - prediction_probabilities,
                    prediction_probabilities,
                ),
                1,
            )
        return prediction_probabilities

    def fit(
        self,
        x: NDFloat,
        y: NDNumber,
        x_val: Optional[NDFloat] = None,
        y_val: Optional[NDNumber] = None,
        num_workers: int = 2,
        custom_train_dataloader:DataLoader=None,
        custom_val_dataloader:DataLoader=None,
    ) -> None:
        self.model.fit(x, y)

    def clone(self) -> Model:
        return self.__class__(**self.constructor_kwargs)

    @abc.abstractmethod
    def load(self, path: str) -> None:
        pass

    @abc.abstractmethod
    def save(self, path: str) -> None:
        pass



from __future__ import annotations

import typing
from typing import Any, List, Union


def _check_min_operands_length(
    expected: int, operands: List[typing.Any]
) -> None:
    if len(operands) < expected:
        raise ValueError(
            f"Operands length={len(operands)}, expected: >= {expected}."
        )


class ConstraintsNode:
    def accept(self, visitor: typing.Any) -> typing.Any:
        return visitor.visit(self)


# ------------ Values
class Value(ConstraintsNode):
    def __add__(self, other: Value) -> ManySum:
        if isinstance(self, ManySum):
            return ManySum(self.operands + [other])
        elif isinstance(other, ManySum):
            return ManySum([self] + other.operands)
        else:
            return ManySum([self, other])

    def __sub__(self, other: Value) -> MathOperation:
        return MathOperation("-", self, other)

    def __mul__(self, other: Value) -> MathOperation:
        return MathOperation("*", self, other)

    def __truediv__(self, other: Value) -> MathOperation:
        return MathOperation("/", self, other)

    def __pow__(self, power: Value, modulo: Any = None) -> MathOperation:
        return MathOperation("**", self, power)

    def __mod__(self, other: Value) -> MathOperation:
        return MathOperation("%", self, other)

    def __lt__(self, other: Value) -> BaseRelationConstraint:
        return LessConstraint(self, other)

    def __le__(self, other: Value) -> BaseRelationConstraint:
        return LessEqualConstraint(self, other)

    def __eq__(self, other: object) -> typing.Any:
        if isinstance(other, Value):
            return EqualConstraint(self, other)
        else:
            print(type(other))
            raise ValueError


class Constant(Value):
    def __init__(self, constant: Union[int, float]):
        self.constant = constant


class Feature(Value):
    def __init__(self, feature_id: Union[str, int]):
        self.feature_id = feature_id


# ------------
class MathOperation(Value):
    def __init__(
        self, operator: str, left_operand: Value, right_operand: Value
    ):
        self.operator = operator
        self.left_operand = left_operand
        self.right_operand = right_operand


class SafeDivision(Value):
    def __init__(self, dividend: Value, divisor: Value, fill_value: Value):
        self.dividend = dividend
        self.divisor = divisor
        self.fill_value = fill_value


class ManySum(Value):
    def __init__(self, operands: List[Value]) -> None:
        self.operands = operands


class Log(Value):
    def __init__(self, operand: Value, safe_value: Value = None):
        self.operand = operand
        self.safe_value = safe_value


# ------------ Constraints


class BaseRelationConstraint(ConstraintsNode):
    def __or__(self, other: BaseRelationConstraint) -> OrConstraint:
        return OrConstraint([self, other])

    def __and__(self, other: BaseRelationConstraint) -> AndConstraint:
        return AndConstraint([self, other])


# ------ Binary


class OrConstraint(BaseRelationConstraint):
    def __init__(self, operands: List[BaseRelationConstraint]):
        _check_min_operands_length(2, operands)
        self.operands = operands


class AndConstraint(BaseRelationConstraint):
    def __init__(self, operands: List[BaseRelationConstraint]):
        _check_min_operands_length(2, operands)
        self.operands = operands


# ----- Comparison


class LessEqualConstraint(BaseRelationConstraint):
    def __init__(self, left_operand: Value, right_operand: Value):
        self.left_operand = left_operand
        self.right_operand = right_operand


class LessConstraint(BaseRelationConstraint):
    def __init__(self, left_operand: Value, right_operand: Value):
        self.left_operand = left_operand
        self.right_operand = right_operand


class EqualConstraint(BaseRelationConstraint):
    def __init__(self, left_operand: Value, right_operand: Value):
        self.left_operand = left_operand
        self.right_operand = right_operand


# ----- Extension


class Count(Value):
    def __init__(
        self, operands: List[BaseRelationConstraint], inverse: bool = False
    ) -> None:
        _check_min_operands_length(1, operands)
        self.operands = operands
        self.inverse = inverse



import enum
from typing import Any, Optional, Tuple, Dict, Union, cast

import numpy as np
import scipy.special
import sklearn.metrics as skm

from . import util
from .util import TaskType


class PredictionType(enum.Enum):
    LOGITS = 'logits'
    PROBS = 'probs'


class MetricsReport:
    def __init__(self, report: dict, task_type: TaskType):
        self._res = {k: {} for k in report.keys()}
        if task_type in (TaskType.BINCLASS, TaskType.MULTICLASS):
            self._metrics_names = ["acc", "f1"]
            for k in report.keys():
                self._res[k]["acc"] = report[k]["accuracy"]
                self._res[k]["f1"] = report[k]["macro avg"]["f1-score"]
                if task_type == TaskType.BINCLASS:
                    self._res[k]["roc_auc"] = report[k]["roc_auc"]
                    self._metrics_names.append("roc_auc")

        elif task_type == TaskType.REGRESSION:
            self._metrics_names = ["r2", "rmse"]
            for k in report.keys():
                self._res[k]["r2"] = report[k]["r2"]
                self._res[k]["rmse"] = report[k]["rmse"]
        else:
            raise "Unknown TaskType!"

    def get_splits_names(self) -> list[str]:
        return self._res.keys()

    def get_metrics_names(self) -> list[str]:
        return self._metrics_names

    def get_metric(self, split: str, metric: str) -> float:
        return self._res[split][metric]

    def get_val_score(self) -> float:
        return self._res["val"]["r2"] if "r2" in self._res["val"] else self._res["val"]["f1"]

    def get_test_score(self) -> float:
        return self._res["test"]["r2"] if "r2" in self._res["test"] else self._res["test"]["f1"]

    def print_metrics(self) -> None:
        res = {
            "val": {k: np.around(self._res["val"][k], 4) for k in self._res["val"]},
            "test": {k: np.around(self._res["test"][k], 4) for k in self._res["test"]}
        }

        print("*" * 100)
        print("[val]")
        print(res["val"])
        print("[test]")
        print(res["test"])

        return res


class SeedsMetricsReport:
    def __init__(self):
        self._reports = []

    def add_report(self, report: MetricsReport) -> None:
        self._reports.append(report)

    def get_mean_std(self) -> dict:
        res = {k: {} for k in ["train", "val", "test"]}
        for split in self._reports[0].get_splits_names():
            for metric in self._reports[0].get_metrics_names():
                res[split][metric] = [x.get_metric(split, metric) for x in self._reports]

        agg_res = {k: {} for k in ["train", "val", "test"]}
        for split in self._reports[0].get_splits_names():
            for metric in self._reports[0].get_metrics_names():
                for k, f in [("count", len), ("mean", np.mean), ("std", np.std)]:
                    agg_res[split][f"{metric}-{k}"] = f(res[split][metric])
        self._res = res
        self._agg_res = agg_res

        return agg_res

    def print_result(self) -> dict:
        res = {split: {k: float(np.around(self._agg_res[split][k], 4)) for k in self._agg_res[split]} for split in
               ["val", "test"]}
        print("=" * 100)
        print("EVAL RESULTS:")
        print("[val]")
        print(res["val"])
        print("[test]")
        print(res["test"])
        print("=" * 100)
        return res


def calculate_rmse(
        y_true: np.ndarray, y_pred: np.ndarray, std: Optional[float]
) -> float:
    rmse = skm.mean_squared_error(y_true, y_pred) ** 0.5
    if std is not None:
        rmse *= std
    return rmse


def _get_labels_and_probs(
        y_pred: np.ndarray, task_type: TaskType, prediction_type: Optional[PredictionType]
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    assert task_type in (TaskType.BINCLASS, TaskType.MULTICLASS)

    if prediction_type is None:
        return y_pred, None

    if prediction_type == PredictionType.LOGITS:
        probs = (
            scipy.special.expit(y_pred)
            if task_type == TaskType.BINCLASS
            else scipy.special.softmax(y_pred, axis=1)
        )
    elif prediction_type == PredictionType.PROBS:
        probs = y_pred
    else:
        util.raise_unknown('prediction_type', prediction_type)

    assert probs is not None
    labels = np.round(probs) if task_type == TaskType.BINCLASS else probs.argmax(axis=1)
    return labels.astype('int64'), probs


def calculate_metrics(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        task_type: Union[str, TaskType],
        prediction_type: Optional[Union[str, PredictionType]],
        y_info: Dict[str, Any],
) -> Dict[str, Any]:
    # Example: calculate_metrics(y_true, y_pred, 'binclass', 'logits', {})
    task_type = TaskType(task_type)
    if prediction_type is not None:
        prediction_type = PredictionType(prediction_type)

    if task_type == TaskType.REGRESSION:
        assert prediction_type is None
        assert 'std' in y_info
        rmse = calculate_rmse(y_true, y_pred, y_info['std'])
        r2 = skm.r2_score(y_true, y_pred)
        result = {'rmse': rmse, 'r2': r2}
    else:
        labels, probs = _get_labels_and_probs(y_pred, task_type, prediction_type)
        result = cast(
            Dict[str, Any], skm.classification_report(y_true, labels, output_dict=True)
        )
        if task_type == TaskType.BINCLASS:
            result['roc_auc'] = skm.roc_auc_score(y_true, probs)
    return result



import os
import sys

sys.path.append(".")
from mlc.logging import XP

import torch
import optuna
from optuna.trial import TrialState
from mlc.datasets.dataset_factory import load_dataset
from mlc.models.model_factory import load_model
from argparse import ArgumentParser
from mlc.transformers.tab_scaler import TabScaler
from mlc.metrics.compute import compute_metric
from mlc.metrics.metric_factory import create_metric
from mlc.dataloaders import get_custom_dataloader

# Torch config to avoid crash on HPC
torch.multiprocessing.set_sharing_strategy("file_system")

CUSTOM_DATALOADERS = ["default", "subset", "madry"]


def run(
    dataset_name: str = "lcld_v2_iid",
    model_name: str = "tabtransformer",
    n_trials: int = 100,
    subset: int = 1000,
    train_batch_size: int = 256,
    val_batch_size: int = 1024,
    epochs: int = 100,
    verbose: int = 0,
    device="cuda",
    custom_dataloaders: list = CUSTOM_DATALOADERS,
) -> None:
    print(
        "Train hyperparameter optimization for {} on {}".format(
            model_name, dataset_name
        )
    )
    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    common_model_params = {
        "x_metadata": metadata,
        "objective": "classification",
        "use_gpu": True,
        "batch_size": train_batch_size,
        "num_classes": 2,
        "early_stopping_rounds": 5,
        "val_batch_size": val_batch_size,
        "class_weight": "balanced",
        "custom_dataloader": "",
    }

    x, y = dataset.get_x_y()
    splits = dataset.get_splits()

    x_train = x.iloc[splits["train"]]
    y_train = y[splits["train"]]

    if subset > 0:
        x_train = x_train[:subset]
        y_train = y_train[:subset]

    model_class = load_model(model_name)
    args = {
        **common_model_params,
        "depth": 12,
        "heads": 2,  # 4,
        "weight_decay": -3,  # -2,
        "learning_rate": -3,  # -2,
        "dropout": 0.5,  # 0.3,
        "data_parallel": False,
        "epochs": epochs,
        "model_name": model_name,
        "dataset": dataset_name,
        "early_stopping_rounds": 5,
        "num_splits": 5,
        "seed": 42,
        "shuffle": True,
        "metrics": ["auc"],
    }
    if model_name == "torchrln":
        args["weight_decay"] = 0

    if not (model_name == "vime" and dataset_name == "malware"):

        study_name = model_name + "_" + dataset_name
        storage_name = "sqlite:///{}.db".format(study_name)

        study = optuna.create_study(
            direction="maximize",
            study_name=study_name,
            storage=storage_name,
            load_if_exists=True,
        )

        n_completed = len(study.get_trials(states=(TrialState.COMPLETE,)))
        print(f"{n_completed}/{n_trials} already completed")
        best_args = {
            **args,
            **study.best_trial.params,
            "task": "train",
            "trial": "best",
        }
        best_args["x_metadata"] = None
        best_trial = sorted(
            study.trials, key=lambda d: d.value if d.values is not None else -1
        )[-1]
        print(
            f"Best trial parameters: {best_trial.params} with best performance: {best_trial.value}"
        )
    else:
        best_args = {
            **args,
            "task": "train",
            "trial": "best",
        }

    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )

    if model_name == "vime" and dataset_name == "malware":
        best_trial = {}

    for custom_dataloader in custom_dataloaders:

        if model_name == "vime" and dataset_name == "malware":
            best_args = {
                **best_args,
                "custom_dataloader": custom_dataloader,
            }
        else:
            best_args = {
                **best_args,
                **best_trial.params,
                "best_trial_value": best_trial.value,
                "best_trial_index": best_trial.number,
                "custom_dataloader": custom_dataloader,
            }
        experiment = XP(best_args, project_name="mlc_best")
        if model_name == "vime" and dataset_name == "malware":
            model = model_class(
                **{
                    **args,
                    "early_stopping_rounds": args["epochs"],
                    "force_device": device if device != "" else None,
                },
                scaler=scaler,
            )
        else:
            model = model_class(
                **{
                    **args,
                    **best_trial.params,
                    "early_stopping_rounds": args["epochs"],
                    "force_device": device if device != "" else None,
                },
                scaler=scaler,
            )

        x_test = x.iloc[splits["test"]]
        y_test = y[splits["test"]]

        x_dataloader = (
            x.iloc[splits["val"]].values
            if custom_dataloader == "dist"
            else x_train.values
        )
        y_dataloader = (
            y[splits["val"]] if custom_dataloader == "dist" else y_train
        )

        custom_train_dataloader = get_custom_dataloader(
            custom_dataloader,
            dataset,
            model,
            scaler,
            {},
            verbose=verbose,
            x=x_dataloader,
            y=y_dataloader,
            train=True,
            batch_size=train_batch_size,
        )
        model.fit(
            x_train,
            y_train,
            x_test,
            y_test,
            custom_train_dataloader=custom_train_dataloader,
        )

        sc = create_metric("auc")
        metric_val = compute_metric(model, sc, x_test, y_test)
        print(metric_val)
        experiment.log_metric("test_{}".format(sc.metric_name), metric_val)

        save_path = os.path.join(
            ".",
            "data",
            "best_models",
            "{}_{}_{}.model".format(
                model_name, dataset_name, custom_dataloader
            ),
        )
        model.save(save_path)
        experiment.log_model(
            "{}_{}_{}.model".format(
                model_name, dataset_name, custom_dataloader
            ),
            save_path,
        )


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Training with Hyper-parameter optimization"
    )
    parser.add_argument(
        "--dataset_name",
        type=str,
        default="lcld_v2_iid",
    )
    parser.add_argument(
        "--model_name",
        type=str,
        default="tabtransformer",
    )
    parser.add_argument("--n_trials", type=int, default=100)
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--subset", type=int, default=0)
    parser.add_argument("--train_batch_size", type=int, default=1024)
    parser.add_argument("--val_batch_size", type=int, default=2048)
    parser.add_argument("--verbose", type=int, default=0)
    parser.add_argument(
        "--device",
        type=str,
        default="cuda",
    )
    parser.add_argument(
        "--custom_dataloaders",
        type=str,
        default="default+subset+madry",
    )

    args = parser.parse_args()

    run(
        dataset_name=args.dataset_name,
        model_name=args.model_name,
        n_trials=args.n_trials,
        subset=args.subset,
        train_batch_size=args.train_batch_size,
        val_batch_size=args.val_batch_size,
        verbose=args.verbose,
        epochs=args.epochs,
        device=args.device,
        custom_dataloaders=args.custom_dataloaders.split("+"),
    )



from inspect import isfunction

import numpy as np
import torch
import torch.nn.functional as F


def normal_kl(mean1, logvar1, mean2, logvar2):
    """
    Compute the KL divergence between two gaussians.

    Shapes are automatically broadcasted, so batches can be compared to
    scalars, among other use cases.
    """
    tensor = None
    for obj in (mean1, logvar1, mean2, logvar2):
        if isinstance(obj, torch.Tensor):
            tensor = obj
            break
    assert tensor is not None, "at least one argument must be a Tensor"

    # Force variances to be Tensors. Broadcasting helps convert scalars to
    # Tensors, but it does not work for torch.exp().
    logvar1, logvar2 = [
        x if isinstance(x, torch.Tensor) else torch.tensor(x).to(tensor)
        for x in (logvar1, logvar2)
    ]

    return 0.5 * (
            -1.0
            + logvar2
            - logvar1
            + torch.exp(logvar1 - logvar2)
            + ((mean1 - mean2) ** 2) * torch.exp(-logvar2)
    )


def approx_standard_normal_cdf(x):
    """
    A fast approximation of the cumulative distribution function of the
    standard normal.
    """
    return 0.5 * (1.0 + torch.tanh(np.sqrt(2.0 / np.pi) * (x + 0.044715 * torch.pow(x, 3))))


def discretized_gaussian_log_likelihood(x, *, means, log_scales):
    """
    Compute the log-likelihood of a Gaussian distribution discretizing to a
    given image.

    :param x: the target images. It is assumed that this was uint8 values,
              rescaled to the range [-1, 1].
    :param means: the Gaussian mean Tensor.
    :param log_scales: the Gaussian log stddev Tensor.
    :return: a tensor like x of log probabilities (in nats).
    """
    assert x.shape == means.shape == log_scales.shape
    centered_x = x - means
    inv_stdv = torch.exp(-log_scales)
    plus_in = inv_stdv * (centered_x + 1.0 / 255.0)
    cdf_plus = approx_standard_normal_cdf(plus_in)
    min_in = inv_stdv * (centered_x - 1.0 / 255.0)
    cdf_min = approx_standard_normal_cdf(min_in)
    log_cdf_plus = torch.log(cdf_plus.clamp(min=1e-12))
    log_one_minus_cdf_min = torch.log((1.0 - cdf_min).clamp(min=1e-12))
    cdf_delta = cdf_plus - cdf_min
    log_probs = torch.where(
        x < -0.999,
        log_cdf_plus,
        torch.where(x > 0.999, log_one_minus_cdf_min, torch.log(cdf_delta.clamp(min=1e-12))),
    )
    assert log_probs.shape == x.shape
    return log_probs


def sum_except_batch(x, num_dims=1):
    '''
    Sums all dimensions except the first.

    Args:
        x: Tensor, shape (batch_size, ...)
        num_dims: int, number of batch dims (default=1)

    Returns:
        x_sum: Tensor, shape (batch_size,)
    '''
    return x.reshape(*x.shape[:num_dims], -1).sum(-1)


def mean_flat(tensor):
    """
    Take the mean over all non-batch dimensions.
    """
    return tensor.mean(dim=list(range(1, len(tensor.shape))))


def ohe_to_categories(ohe, K):
    K = torch.from_numpy(K)
    indices = torch.cat([torch.zeros((1,)), K.cumsum(dim=0)], dim=0).int().tolist()
    res = []
    for i in range(len(indices) - 1):
        res.append(ohe[:, indices[i]:indices[i + 1]].argmax(dim=1))
    return torch.stack(res, dim=1)


def log_1_min_a(a):
    return torch.log(1 - a.exp() + 1e-40)


def log_add_exp(a, b):
    maximum = torch.max(a, b)
    return maximum + torch.log(torch.exp(a - maximum) + torch.exp(b - maximum))


def exists(x):
    return x is not None


def extract(a, t, x_shape):
    b, *_ = t.shape
    t = t.to(a.device)
    out = a.gather(-1, t)
    while len(out.shape) < len(x_shape):
        out = out[..., None]
    return out.expand(x_shape)


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


def log_categorical(log_x_start, log_prob):
    return (log_x_start.exp() * log_prob).sum(dim=1)


def index_to_log_onehot(x, num_classes):
    onehots = []
    for i in range(len(num_classes)):
        onehots.append(F.one_hot(x[:, i], num_classes[i]))

    x_onehot = torch.cat(onehots, dim=1)
    log_onehot = torch.log(x_onehot.float().clamp(min=1e-30))
    return log_onehot


def log_sum_exp_by_classes(x, slices):
    device = x.device
    res = torch.zeros_like(x)
    for ixs in slices:
        res[:, ixs] = torch.logsumexp(x[:, ixs], dim=1, keepdim=True)

    assert x.size() == res.size()

    return res


@torch.jit.script
def log_sub_exp(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    m = torch.maximum(a, b)
    return torch.log(torch.exp(a - m) - torch.exp(b - m)) + m


@torch.jit.script
def sliced_logsumexp(x, slices):
    lse = torch.logcumsumexp(
        torch.nn.functional.pad(x, [1, 0, 0, 0], value=-float('inf')),
        dim=-1)

    slice_starts = slices[:-1]
    slice_ends = slices[1:]

    slice_lse = log_sub_exp(lse[:, slice_ends], lse[:, slice_starts])
    slice_lse_repeated = torch.repeat_interleave(
        slice_lse,
        slice_ends - slice_starts,
        dim=-1
    )
    return slice_lse_repeated


def log_onehot_to_index(log_x):
    return log_x.argmax(1)


class FoundNANsError(BaseException):
    """Found NANs during sampling"""

    def __init__(self, message='Found NANs during sampling.'):
        super(FoundNANsError, self).__init__(message)



"""
Have not used in TabDDPM project.
"""

import datetime
import shutil
import typing as ty
from pathlib import Path

PROJ = Path('tab-ddpm/').absolute().resolve()
EXP = PROJ / 'exp'
DATA = PROJ / 'data'


def get_path(path: ty.Union[str, Path]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if not path.is_absolute():
        path = PROJ / path
    return path.resolve()


def get_relative_path(path: ty.Union[str, Path]) -> Path:
    return get_path(path).relative_to(PROJ)


def duplicate_path(
        src: ty.Union[str, Path], alternative_project_dir: ty.Union[str, Path]
) -> None:
    src = get_path(src)
    alternative_project_dir = get_path(alternative_project_dir)
    dst = alternative_project_dir / src.relative_to(PROJ)
    dst.parent.mkdir(parents=True, exist_ok=True)
    if dst.exists():
        dst = dst.with_name(
            dst.name + '_' + datetime.datetime.now().strftime('%Y%m%dT%H%M%S')
        )
    (shutil.copytree if src.is_dir() else shutil.copyfile)(src, dst)



# -*- coding:utf-8 -*-
"""

Author:
    Weichen Shen,weichenswc@163.com

"""
import numpy as np
import torch


def concat_fun(inputs, axis=-1):
    if len(inputs) == 1:
        return inputs[0]
    else:
        return torch.cat(inputs, dim=axis)


def slice_arrays(arrays, start=None, stop=None):
    """Slice an array or list of arrays.

    This takes an array-like, or a list of
    array-likes, and outputs:
        - arrays[start:stop] if `arrays` is an array-like
        - [x[start:stop] for x in arrays] if `arrays` is a list

    Can also work on list/array of indices: `slice_arrays(x, indices)`

    Arguments:
        arrays: Single array or list of arrays.
        start: can be an integer index (start index)
            or a list/array of indices
        stop: integer (stop index); should be None if
            `start` was a list.

    Returns:
        A slice of the array(s).

    Raises:
        ValueError: If the value of start is a list and stop is not None.
    """

    if arrays is None:
        return [None]

    if isinstance(arrays, np.ndarray):
        arrays = [arrays]

    if isinstance(start, list) and stop is not None:
        raise ValueError('The stop argument has to be None if the value of start '
                         'is a list.')
    elif isinstance(arrays, list):
        if hasattr(start, '__len__'):
            # hdf5 datasets only support list objects as indices
            if hasattr(start, 'shape'):
                start = start.tolist()
            return [None if x is None else x[start] for x in arrays]
        else:
            if len(arrays) == 1:
                return arrays[0][start:stop]
            return [None if x is None else x[start:stop] for x in arrays]
    else:
        if hasattr(start, '__len__'):
            if hasattr(start, 'shape'):
                start = start.tolist()
            return arrays[start]
        elif hasattr(start, '__getitem__'):
            return arrays[start:stop]
        else:
            return [None]



from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DateSorter,
    DefaultSplitter,
    Task,
)


def create_dataset() -> Dataset:
    dataset = Dataset(
        name="electricity",
        data_source=CsvDataSource(
            path="./data/mlc/electricity/electricity.csv"
        ),
        metadata_source=CsvDataSource(
            path="./data/mlc/electricity/electricity_metadata.csv"
        ),
        tasks=[
            Task(
                name="price_up",
                task_type="classification",
                evaluation_metric="f1_score",
            )
        ],
        sorter=DateSorter(date_col="date_time"),
        splitter=DefaultSplitter(),
        relation_constraints=[],
    )
    return dataset


datasets = [
    {
        "name": "electricity",
        "fun_create": create_dataset,
    },
]



import torch

from mlc.datasets.dataset_factory import load_dataset
from mlc.metrics.compute import compute_metrics
from mlc.metrics.metric_factory import create_metric
from mlc.models.model_factory import load_model
from mlc.transformers.tab_scaler import TabScaler


models_params = {
    "tabtransformer": {
        "model_name": "TabTransformer",
        "dim": 8,
        "batch_size": 64,
        # "num_classes": 1,
        "depth": 6,
        "heads": 4,
        "weight_decay": -2,
        "learning_rate": -3,
        "dropout": 0.1,
        "data_parallel": False,
    },
    "deepfm": {
        "dnn_dropout": 0.3,
        "num_dense_features": None,
    },
    "torchrln": {
        "n_layers": 5,
        "theta": -12,
        "learning_rate": 0.0005588211696108453,
        "norm": 2,
        "hidden_dim": 78,
    },
    "saint": {},
    "vime": {},
}


def run(
    dataset_name: str = "lcld_v2_time",
    model_name: str = "deepfm",
) -> None:

    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    common_model_params = {
        "x_metadata": metadata,
        "objective": "classification",
        "use_gpu": True,
        "batch_size": 64,
        "num_classes": 2,
        "epochs": 1,
        "early_stopping_rounds": 5,
        "val_batch_size": 1024,
        "class_weight": "balanced",
    }

    x, y = dataset.get_x_y()
    splits = dataset.get_splits()

    x_train = x.iloc[splits["train"]]
    y_train = y[splits["train"]]
    x_test = x.iloc[splits["test"]]
    y_test = y[splits["test"]]
    x_val = x.iloc[splits["val"]]
    y_val = y[splits["val"]]

    x_train = x_train.iloc[:200000]
    y_train = y_train[:200000]
    x_test = x_test.iloc[:50000]
    y_test = y_test[:50000]
    x_val = x_val[:1000]
    y_val = y_val[:1000]

    # x_train = x.iloc[:350000]
    # y_train = y[:350000]
    # x_test = x.iloc[400000:]
    # y_test = y[400000:]
    # x_val = x.iloc[350000:400000]
    # y_val = y[350000:400000]

    model_class = load_model(model_name)

    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )

    model_params = {
        **models_params[model_name],
        **common_model_params,
    }
    model = model_class(**model_params, scaler=scaler)

    save_path = f"./data/models/{dataset_name}_{model_name}.model"

    model.fit(x_train.values, y_train, x_val.values, y_val)

    model.save(save_path)
    model = model_class.load_class(
        save_path, x_metadata=metadata, scaler=scaler
    )

    metrics = ["auc", "accuracy", "mcc"]
    metrics_obj = [create_metric(m) for m in metrics]
    res = compute_metrics(model, metrics_obj, x_test.values, y_test)

    for i, e in enumerate(metrics):
        print(f"Test {e}: {res[i]}")


if __name__ == "__main__":
    for e in ["tabtransformer", "deepfm", "torchrln", "saint", "vime"]:
        run(model_name=e)



import os
import sys

sys.path.append(".")
from mlc.logging import XP

import torch
import optuna
from optuna.trial import TrialState
from mlc.datasets.dataset_factory import load_dataset
from mlc.models.model_factory import load_model
from argparse import ArgumentParser, Namespace
from mlc.transformers.tab_scaler import TabScaler
from mlc.metrics.compute import compute_metric
from mlc.metrics.metric_factory import create_metric
from mlc.dataloaders import get_custom_dataloader

# Torch config to avoid crash on HPC
torch.multiprocessing.set_sharing_strategy('file_system')

CUSTOM_DATALOADERS = ["default", "subset", "madry"]


def run(dataset_name: str = "lcld_v2_iid", model_name: str = "tabtransformer", n_trials: int = 100, subset: int = 1000,
        train_batch_size: int = 256, val_batch_size: int = 1024, epochs: int = 100, verbose: int = 0, device="cuda",
        custom_dataloaders: list = CUSTOM_DATALOADERS) -> None:
    print("Train hyperparameter optimization for {} on {}".format(model_name, dataset_name))
    dataset = load_dataset(dataset_name)
    metadata = dataset.get_metadata(only_x=True)
    common_model_params = {
        "x_metadata": metadata,
        "objective": "classification",
        "use_gpu": True,
        "batch_size": train_batch_size,
        "num_classes": 2,
        "early_stopping_rounds": 5,
        "val_batch_size": val_batch_size,
        "class_weight": "balanced",
        "custom_dataloader": ""
    }

    x, y = dataset.get_x_y()
    splits = dataset.get_splits()

    x_train = x.iloc[splits["train"]]
    y_train = y[splits["train"]]

    if subset > 0:
        x_train = x_train[:subset]
        y_train = y_train[:subset]

    model_class = load_model(model_name)
    args = {
        **common_model_params,
        "depth": 12,
        "heads": 2,  # 4,
        "weight_decay": -3,  # -2,
        "learning_rate": -3,  # -2,
        "dropout": 0.5,  # 0.3,
        "data_parallel": False,
        "epochs": epochs,
        "model_name": model_name,
        "dataset": dataset_name,
        "early_stopping_rounds": 5,
        "num_splits": 5,
        "seed": 42,
        "shuffle": True,
        "metrics": ["auc"]
    }
    if model_name == "torchrln":
        args["weight_decay"] = 0

    study_name = model_name + "_" + dataset_name
    storage_name = "sqlite:///{}.db".format(study_name)

    study = optuna.create_study(direction="maximize",
                                study_name=study_name,
                                storage=storage_name,
                                load_if_exists=True)

    n_completed = len(study.get_trials(states=(TrialState.COMPLETE,)))
    print(f"{n_completed}/{n_trials} already completed")

    scaler = TabScaler(num_scaler="min_max", one_hot_encode=True)
    scaler.fit(
        torch.tensor(x.values, dtype=torch.float32), x_type=metadata["type"]
    )

    best_args = {**args, "task": "train", "trial": "best"}
    if len(study.trials) > 1:
        best_trial = sorted(study.trials, key=lambda d: d.value if d.values is not None else -1)[-1]
        print(f"Best trial parameters: {best_trial.params} with best performance: {best_trial.value}")
        best_args = {**best_args, **study.best_trial.params}
    else:
        best_trial = Namespace(params={},value=0, number=-1)
        print("There is no best trial, using default params!!!")

    best_args["x_metadata"] = None

    for custom_dataloader in custom_dataloaders:
        best_args = {**best_args, **best_trial.params, "best_trial_value": best_trial.value,
                     "best_trial_index": best_trial.number, "custom_dataloader": custom_dataloader}
        experiment = XP(best_args, project_name="mlc_best_v2")

        model = model_class(**{**args, **best_trial.params, "early_stopping_rounds": args["epochs"],
                               "force_device":device if device!="" else None}, scaler=scaler)

        x_test = x.iloc[splits["test"]]
        y_test = y[splits["test"]]

        x_dataloader = x.iloc[splits["val"]].values if custom_dataloader=="dist" else x_train.values
        y_dataloader = y[splits["val"]] if custom_dataloader == "dist" else y_train

        custom_train_dataloader = get_custom_dataloader(custom_dataloader, dataset, model, scaler, {}, verbose=verbose,
                                                        x=x_dataloader, y=y_dataloader, train=True,
                                                        batch_size=train_batch_size)
        model.fit(x_train, y_train, x_test, y_test, custom_train_dataloader=custom_train_dataloader)

        sc = create_metric("auc")
        metric_val = compute_metric(model, sc, x_test, y_test)
        print(metric_val)
        experiment.log_metric("test_{}".format(sc.metric_name), metric_val)

        save_path = os.path.join(".", "data", "best_models", "{}_{}_{}.model".format(model_name, dataset_name,
                                                                                     custom_dataloader))
        model.save(save_path)
        experiment.log_model("{}_{}_{}.model".format(model_name, dataset_name, custom_dataloader), save_path)


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Training with Hyper-parameter optimization"
    )
    parser.add_argument("--dataset_name", type=str, default="lcld_v2_iid",
                        )
    parser.add_argument("--model_name", type=str, default="tabtransformer",
                        )
    parser.add_argument("--n_trials", type=int, default=100)
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--subset", type=int, default=0)
    parser.add_argument("--train_batch_size", type=int, default=1024)
    parser.add_argument("--val_batch_size", type=int, default=2048)
    parser.add_argument("--verbose", type=int, default=0)
    parser.add_argument("--device", type=str, default="cuda",
                        )
    parser.add_argument("--custom_dataloaders", type=str, default="default+subset+madry",
                        )

    args = parser.parse_args()

    run(dataset_name=args.dataset_name, model_name=args.model_name, n_trials=args.n_trials, subset=args.subset,
        train_batch_size=args.train_batch_size, val_batch_size=args.val_batch_size, verbose=args.verbose,
        epochs=args.epochs, device=args.device, custom_dataloaders=args.custom_dataloaders.split("+"))



import json
import os.path
from pathlib import Path
from types import FunctionType
from typing import Any, Dict, Optional

import h5py
import joblib
import numpy as np
import numpy.typing as npt
from sklearn.base import BaseEstimator

cache = []
hdf5_dataset_key = "default"


def load_json(path: str) -> Dict[str, Any]:
    with open(path, "r") as f:
        obj = json.load(f)
    return obj


def save_json(obj: Any, path: str) -> None:
    with open(path, "w") as f:
        json.dump(obj, f, indent=4)


def load_hdf5(path: str) -> npt.NDArray[Any]:

    with h5py.File(path, "r") as f:
        return f[hdf5_dataset_key][()]


def save_hdf5(obj: Any, path: str) -> None:
    with h5py.File(path, "w") as f:
        f.create_dataset(hdf5_dataset_key, data=obj, compression="gzip")


type_to_fun_mapper = [
    {
        "type": np.ndarray,
        "load": load_hdf5,
        "save": save_hdf5,
    },
    {"type": BaseEstimator, "load": joblib.load, "save": joblib.dump},
    {"type": dict, "load": load_json, "save": save_json},
]


def find_mapper(out_type: Any) -> Dict[str, Any]:
    for mapper in type_to_fun_mapper:
        mapper_type = mapper.get("type")
        if issubclass(out_type, mapper_type):
            return mapper
    raise NotImplementedError


def load_do_save(
    path: str,
    executable: FunctionType,
    return_type: Optional[str] = None,
    verbose: bool = False,
    **kwargs: Dict[str, Any],
) -> Any:
    if return_type is None:
        return_type = executable.__annotations__["return"]

    mapper = find_mapper(return_type)
    if os.path.exists(path):
        load = mapper["load"]
        out = load(path)
        if verbose:
            print(f"{path} loaded.")
        return out
    else:
        save = mapper["save"]
        obj = executable(**kwargs)
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        save(obj, path)
        if verbose:
            print(f"{path} saved.")
        return obj



# -*- coding:utf-8 -*-
"""
Author:
    Weichen Shen,weichenswc@163.com
Reference:
    [1] Guo H, Tang R, Ye Y, et al. Deepfm: a factorization-machine based neural network for ctr prediction[J]. arXiv preprint arXiv:1703.04247, 2017.(https://arxiv.org/abs/1703.04247)
"""
import torch
import torch.nn as nn

from .basemodel import BaseModel
from ..inputs import combined_dnn_input
from ..layers import FM, DNN


class DeepFM(BaseModel):
    """Instantiates the DeepFM Network architecture.

    :param linear_feature_columns: An iterable containing all the features used by linear part of the model.
    :param dnn_feature_columns: An iterable containing all the features used by deep part of the model.
    :param use_fm: bool,use FM part or not
    :param dnn_hidden_units: list,list of positive integer or empty list, the layer number and units in each layer of DNN
    :param l2_reg_linear: float. L2 regularizer strength applied to linear part
    :param l2_reg_embedding: float. L2 regularizer strength applied to embedding vector
    :param l2_reg_dnn: float. L2 regularizer strength applied to DNN
    :param init_std: float,to use as the initialize std of embedding vector
    :param seed: integer ,to use as random seed.
    :param dnn_dropout: float in [0,1), the probability we will drop out a given DNN coordinate.
    :param dnn_activation: Activation function to use in DNN
    :param dnn_use_bn: bool. Whether use BatchNormalization before activation or not in DNN
    :param task: str, ``"binary"`` for  binary logloss or  ``"regression"`` for regression loss
    :param device: str, ``"cpu"`` or ``"cuda:0"``
    :param gpus: list of int or torch.device for multiple gpus. If None, run on `device`. `gpus[0]` should be the same gpu with `device`.
    :return: A PyTorch model instance.

    """

    def __init__(self,
                 linear_feature_columns, dnn_feature_columns, use_fm=True,
                 dnn_hidden_units=(256, 128),
                 l2_reg_linear=0.00001, l2_reg_embedding=0.00001, l2_reg_dnn=0, init_std=0.0001, seed=1024,
                 dnn_dropout=0,
                 dnn_activation='relu', dnn_use_bn=False, task='binary', device='cpu', gpus=None):

        super(DeepFM, self).__init__(linear_feature_columns, dnn_feature_columns, l2_reg_linear=l2_reg_linear,
                                     l2_reg_embedding=l2_reg_embedding, init_std=init_std, seed=seed, task=task,
                                     device=device, gpus=gpus)

        self.use_fm = use_fm
        self.use_dnn = len(dnn_feature_columns) > 0 and len(
            dnn_hidden_units) > 0
        if use_fm:
            self.fm = FM()

        if self.use_dnn:
            self.dnn = DNN(self.compute_input_dim(dnn_feature_columns), dnn_hidden_units,
                           activation=dnn_activation, l2_reg=l2_reg_dnn, dropout_rate=dnn_dropout, use_bn=dnn_use_bn,
                           init_std=init_std, device=device)
            self.dnn_linear = nn.Linear(
                dnn_hidden_units[-1], 1, bias=False).to(device)

            self.add_regularization_weight(
                filter(lambda x: 'weight' in x[0] and 'bn' not in x[0], self.dnn.named_parameters()), l2=l2_reg_dnn)
            self.add_regularization_weight(self.dnn_linear.weight, l2=l2_reg_dnn)
        self.to(device)

    def forward(self, X):

        sparse_embedding_list, dense_value_list = self.input_from_feature_columns(X, self.dnn_feature_columns,
                                                                                  self.embedding_dict)
        logit = self.linear_model(X)

        if self.use_fm and len(sparse_embedding_list) > 0:
            fm_input = torch.cat(sparse_embedding_list, dim=1)
            logit += self.fm(fm_input)

        if self.use_dnn:
            dnn_input = combined_dnn_input(
                sparse_embedding_list, dense_value_list)
            dnn_output = self.dnn(dnn_input)
            dnn_logit = self.dnn_linear(dnn_output)
            logit += dnn_logit

        y_pred = self.out(logit)

        return y_pred



from abc import abstractmethod
from typing import Tuple

import torch
from torch.utils.data import DataLoader


class BaseDataLoader:
    @abstractmethod
    def get_dataloader(
            self, x: torch.Tensor, y: torch.Tensor, train: bool, batch_size: int
    ) -> DataLoader[Tuple[torch.Tensor, ...]]:
        raise NotImplementedError



from abc import abstractmethod
from typing import Any, Union

import numpy as np
import numpy.typing as npt
import torch

from mlc.constraints.backend import Backend
from mlc.constraints.pytorch_backend import PytorchBackend
from mlc.constraints.relation_constraint import (
    AndConstraint,
    BaseRelationConstraint,
    Constant,
    ConstraintsNode,
    Count,
    EqualConstraint,
    Feature,
    LessConstraint,
    LessEqualConstraint,
    Log,
    ManySum,
    MathOperation,
    OrConstraint,
    SafeDivision,
    Value,
)
from mlc.constraints.utils import get_feature_index
from mlc.typing import NDNumber

EPS: npt.NDArray[Any] = np.array(0.000001)


class ConstraintsVisitor:
    """Abstract Visitor Class"""

    @abstractmethod
    def visit(self, item: ConstraintsNode) -> Any:
        pass

    @abstractmethod
    def execute(self) -> Any:
        pass


class BackendConstraintsVisitor(ConstraintsVisitor):
    def __init__(
        self,
        constraint: Union[BaseRelationConstraint, Value],
        x: Union[NDNumber, torch.Tensor],
        backend: Backend,
        feature_names: npt.ArrayLike = None,
    ) -> None:
        self.constraint = constraint
        self.x = x
        self.backend = backend
        self.feature_names = feature_names

    def visit(self, constraint_node: ConstraintsNode) -> Any:

        # ------------ Values
        if isinstance(constraint_node, Constant):
            return self.backend.constant(constraint_node.constant)

        elif isinstance(constraint_node, Feature):
            feature_index = get_feature_index(
                self.feature_names, constraint_node.feature_id
            )
            return self.backend.feature(self.x, feature_index)

        elif isinstance(constraint_node, MathOperation):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self).to(left_operand.device) if isinstance(left_operand,torch.Tensor)  else constraint_node.right_operand.accept(self)
            operator = constraint_node.operator
            return self.backend.math_operation(
                operator, left_operand, right_operand
            )

        elif isinstance(constraint_node, SafeDivision):
            dividend = constraint_node.dividend.accept(self)
            divisor = constraint_node.divisor.accept(self)
            fill_value = constraint_node.fill_value.accept(self)
            return self.backend.safe_division(dividend, divisor, fill_value)

        elif isinstance(constraint_node, Log):
            operand = constraint_node.operand.accept(self)
            if constraint_node.safe_value is not None:
                safe_value = constraint_node.safe_value.accept(self)
            else:
                safe_value = None

            return self.backend.log(operand, safe_value)

        elif isinstance(constraint_node, ManySum):
            operands = [e.accept(self) for e in constraint_node.operands]

            return self.backend.many_sum(operands)

        # ------------ Constraints

        # ------ Binary
        elif isinstance(constraint_node, OrConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            return self.backend.or_constraint(operands)

        elif isinstance(constraint_node, AndConstraint):
            operands = [e.accept(self) for e in constraint_node.operands]
            return self.backend.and_constraint(operands)

        # ------ Comparison
        elif isinstance(constraint_node, LessEqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return self.backend.less_equal_constraint(
                left_operand, right_operand
            )

        elif isinstance(constraint_node, LessConstraint):
            left_operand = constraint_node.left_operand.accept(self) + self.backend.constant(EPS)
            right_operand = constraint_node.right_operand.accept(self)
            return self.backend.less_constraint(left_operand, right_operand)

        elif isinstance(constraint_node, EqualConstraint):
            left_operand = constraint_node.left_operand.accept(self)
            right_operand = constraint_node.right_operand.accept(self)
            return self.backend.equal_constraint(left_operand, right_operand)

            # ------ Extension

        elif isinstance(constraint_node, Count):
            operands = [e.accept(self) for e in constraint_node.operands]
            inverse = constraint_node.inverse
            return self.backend.count(operands, inverse)

        else:
            raise NotImplementedError

    def execute(self) -> torch.Tensor:
        return self.constraint.accept(self)


class ConstraintsExecutor:
    def __init__(
        self,
        constraint: Union[BaseRelationConstraint, Value],
        backend: Backend,
        feature_names: npt.ArrayLike = None,
    ):
        self.constraint = constraint
        self.backend = backend
        self.feature_names = feature_names

    def execute(
        self, x: Union[NDNumber, torch.Tensor]
    ) -> Union[NDNumber, torch.Tensor]:
        if isinstance(self.backend, PytorchBackend):
            self.backend.set_device(x.device)
        
        visitor = BackendConstraintsVisitor(
            self.constraint, x, self.backend, self.feature_names
        )
        return visitor.execute()



from .model import *


class sep_MLP(nn.Module):
    def __init__(self, dim, len_feats, categories):
        super(sep_MLP, self).__init__()
        self.len_feats = len_feats
        self.layers = nn.ModuleList([])
        for i in range(len_feats):
            self.layers.append(simple_MLP([dim, 5 * dim, categories[i]]))

    def forward(self, x):
        y_pred = list([])
        for i in range(self.len_feats):
            x_i = x[:, i, :]
            pred = self.layers[i](x_i)
            y_pred.append(pred)
        return y_pred


class SAINT_vision(nn.Module):
    def __init__(
        self,
        *,
        categories,
        num_continuous,
        dim,
        depth,
        heads,
        dim_head=16,
        dim_out=1,
        mlp_hidden_mults=(4, 2),
        mlp_act=None,
        num_special_tokens=0,
        continuous_mean_std=None,
        attn_dropout=0.0,
        ff_dropout=0.0,
        cont_embeddings="MLP",
        scalingfactor=10,
        attentiontype="col",
        final_mlp_style="common",
        y_dim=2,
    ):
        super().__init__()
        assert all(
            map(lambda n: n > 0, categories)
        ), "number of each category must be positive"

        # categories related calculations

        self.num_categories = len(categories)
        self.num_unique_categories = sum(categories)

        # create category embeddings table

        self.num_special_tokens = num_special_tokens
        self.total_tokens = categories[-1] + 256

        # for automatically offsetting unique category ids to the correct position in the categories embedding table

        categories_offset = torch.tensor(
            np.append(np.repeat(0, self.num_categories - 1), [256])
        )
        self.register_buffer("categories_offset", categories_offset)

        self.norm = nn.LayerNorm(num_continuous)
        self.num_continuous = num_continuous
        self.dim = dim
        self.cont_embeddings = cont_embeddings
        self.attentiontype = attentiontype
        self.final_mlp_style = final_mlp_style

        if self.cont_embeddings == "MLP":
            self.simple_MLP = nn.ModuleList(
                [
                    simple_MLP([1, 100, self.dim])
                    for _ in range(self.num_continuous)
                ]
            )
            input_size = (dim * self.num_categories) + (dim * num_continuous)
            nfeats = self.num_categories + num_continuous
        else:
            print("Continous features are not passed through attention")
            input_size = (dim * self.num_categories) + num_continuous
            nfeats = self.num_categories

            # transformer
        if attentiontype == "col":
            self.transformer = Transformer(
                num_tokens=self.total_tokens,
                dim=dim,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout,
            )
        elif attentiontype in ["row", "colrow"]:
            self.transformer = RowColTransformer(
                num_tokens=self.total_tokens,
                dim=dim,
                nfeats=nfeats,
                depth=depth,
                heads=heads,
                dim_head=dim_head,
                attn_dropout=attn_dropout,
                ff_dropout=ff_dropout,
                style=attentiontype,
            )

        l = input_size // 8
        hidden_dimensions = list(map(lambda t: l * t, mlp_hidden_mults))
        all_dimensions = [input_size, *hidden_dimensions, dim_out]

        self.mlp = MLP(all_dimensions, act=mlp_act)
        self.embeds = nn.Embedding(self.total_tokens, self.dim)

        cat_mask_offset = torch.tensor(
            np.append(np.repeat(0, self.num_categories - 1), [2])
        )
        con_mask_offset = torch.empty(0)

        self.register_buffer("cat_mask_offset", cat_mask_offset)
        self.register_buffer("con_mask_offset", con_mask_offset)

        self.mask_embeds_cat = nn.Embedding(4, self.dim)
        self.mask_embeds_cont = nn.Embedding(4, self.dim)
        self.pos_encodings = nn.Embedding(self.num_categories, self.dim)
        if self.final_mlp_style == "common":
            self.mlp1 = simple_MLP(
                [dim, (self.total_tokens) * 2, self.total_tokens]
            )
            self.mlp2 = simple_MLP([dim, (self.num_continuous), 1])

        else:
            self.mlp1 = sep_MLP(dim, self.num_categories, categories)
            self.mlp2 = sep_MLP(
                dim,
                self.num_continuous,
                np.ones(self.num_continuous).astype(int),
            )

        self.mlpfory = simple_MLP([dim, 100, y_dim])

    def forward(self, x_categ, x_cont):
        x = self.transformer(x_categ, x_cont)
        y_reps = x[:, self.num_categories - 1, :]
        y_outs = self.mlpfory(y_reps)
        return y_outs



from typing import List, Tuple, Type, Union

from mlc.models.model import Model
from mlc.models.samples.lcld import models as lcld_models
from mlc.models.samples.sklearn_default import models as sk_models
from mlc.models.tabsurvey import models as tabsurvey_models
from mlc.models.samples.random_forest import models as rf_models

models: List[Tuple[str, Type[Model]]] = (
    sk_models + lcld_models + tabsurvey_models + rf_models
)


def load_model(model_name: str) -> Type[Model]:
    return load_models(model_name)[0]


def load_models(model_names: Union[str, List[str]]) -> List[Type[Model]]:

    if isinstance(model_names, str):
        model_names = [model_names]
    models_out = list(filter(lambda e: e[0] in model_names, models))
    models_out = [e[1] for e in models_out]

    if len(models_out) != len(model_names):
        raise NotImplementedError("At least one model is not available.")

    return models_out



from typing import List, Union

import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, StandardScaler


def get_numeric_categorical_preprocessor(
    numeric_features: Union[List[str], pd.Series],
    categorical_features: Union[List[str], pd.Series],
) -> ColumnTransformer:

    numeric_transformer = Pipeline(
        steps=[
            ("imputer", SimpleImputer(strategy="median")),
            ("scaler", StandardScaler()),
        ]
    )
    categorical_transformer = OneHotEncoder(handle_unknown="ignore")

    preprocessor = ColumnTransformer(
        transformers=[
            ("num", numeric_transformer, numeric_features),
            ("cat", categorical_transformer, categorical_features),
        ],
        sparse_threshold=0,
        n_jobs=-1,
    )

    return preprocessor



import importlib
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from mlc.constraints.relation_constraint import BaseRelationConstraint
from mlc.datasets.dataset import (
    CsvDataSource,
    Dataset,
    DataSource,
    DateSorter,
    DownloadFileDataSource,
    FileDataSource,
    Hdf5DataSource,
    SQLDataSource,
    Task,
)
from mlc.datasets.samples import load_dataset


def get_tasks(config: List[Dict[str, Any]]) -> List[Task]:
    return [Task(**e) for e in config]


def get_sorter(config: Dict[str, Any]) -> Optional[DateSorter]:
    if "date" in config.keys():
        return DateSorter(config.get("date"))
    else:
        return None


extension_to_file_loader = {".csv": CsvDataSource, ".hdf5": Hdf5DataSource}


def get_filedatasource(path: str) -> FileDataSource:
    ext = Path(path).suffix
    if ext not in extension_to_file_loader:
        raise NotImplementedError
    return extension_to_file_loader[ext](path)


def get_fun_preprocess(module_name: str) -> Optional[Any]:
    return importlib.import_module(module_name).preprocess


def get_sql_datasource(config: Dict[str, Any]) -> DataSource:

    fun_preprocess = None
    if config.get("preprocess_module") is not None:
        fun_preprocess = get_fun_preprocess(config["preprocess_module"])

    return SQLDataSource(
        config["connection_string"],
        config["query"],
        fun_preprocess,
    )


def get_datasource(config: Dict[str, Any]) -> DataSource:
    if "download" in config.keys():
        download = config["download"]

        return DownloadFileDataSource(
            url=download.get("url"),
            file_data_source=get_filedatasource(download.get("cache")),
        )
    if "file" in config.keys():
        file = config["file"]
        return get_filedatasource(file.get("path"))

    if "sql" in config.keys():
        return get_sql_datasource(config["sql"])


def get_dataset(config: Union[Dict[str, Any], str]) -> Dataset:
    if isinstance(config, str):
        config = {"name": config}
    return get_dataset_from_config(config)


def get_dataset_from_config(config: Dict[str, Any]) -> Dataset:

    name = config.get("name")

    # If it is a known dataset
    if len(config.keys()) == 1:
        return load_dataset(name)

    data_source = get_datasource(config["data"].get("source"))
    metadata_source = get_datasource(config["metadata"].get("source"))
    tasks = get_tasks(config["tasks"])
    sorter = get_sorter(config["sorter"])
    splitter = None
    relation_constraints: List[BaseRelationConstraint] = []
    return Dataset(
        name=name,
        data_source=data_source,
        metadata_source=metadata_source,
        tasks=tasks,
        sorter=sorter,
        splitter=splitter,
        relation_constraints=relation_constraints,
    )



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Any, Dict, Optional, Union
from uuid import uuid4

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.optim as optim
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, TensorDataset

from mlc.dataloaders.default import DefaultDataLoader
from mlc.load_do_save import load_json, save_json
from mlc.models.model import Model
from mlc.typing import NDFloat, NDNumber
from mlc.utils import parent_exists

logger = logging.getLogger(__name__)
from mlc.dataloaders.fast_dataloader import FastTensorDataLoader


class BaseModelTorch(Model):
    def __init__(
        self,
        name: str,
        objective: str,
        batch_size: int,
        epochs: int,
        early_stopping_rounds: int,
        learning_rate: float = 0.001,
        val_batch_size: int = 100000,
        class_weight: Union[str, Dict[int, str]] = None,
        force_device: str = None,
        is_text: bool = False,
        weight_decay=0,
        use_gpu: bool = False,
        gpu_ids: Optional[str] = None,
        data_parallel=False,
        **kwargs: Any,
    ) -> None:

        # Generate super call with all the parameters
        super().__init__(
            name=name,
            objective=objective,
            batch_size=batch_size,
            epochs=epochs,
            early_stopping_rounds=early_stopping_rounds,
            learning_rate=learning_rate,
            val_batch_size=val_batch_size,
            class_weight=class_weight,
            weight_decay=weight_decay,
            force_device=force_device,
            is_text=is_text,
            **kwargs,
        )

        self.force_device = force_device
        self.device = self.get_device()
        self.objective = objective
        self.batch_size = batch_size
        self.epochs = epochs
        self.early_stopping_rounds = early_stopping_rounds
        self.learning_rate = learning_rate
        self.force_train = False
        self.val_batch_size = val_batch_size
        self.class_weight = class_weight
        self.is_text = is_text
        self.weight_decay = weight_decay

        # Computed

        self.gpus = (
            gpu_ids
            if use_gpu and torch.cuda.is_available() and data_parallel
            else None
        )

        self.data_loader = DefaultDataLoader()

        if self.objective == "regression":
            self.set_loss(nn.MSELoss())
        elif self.objective == "classification":
            self.set_loss(nn.CrossEntropyLoss())
        else:
            self.set_loss(nn.BCEWithLogitsLoss())

        self.uuid = str(uuid4())

    def to_device(self, device=None) -> None:
        self.device = device if device is not None else self.device
        logger.debug(f"On Device: {self.device}")
        self.model.to(self.device)
        self.data_loader.device = self.device

    def get_device(self):
        if self.force_device is not None:
            device = self.force_device
        else:
            if torch.cuda.is_available():
                device = "cuda"
            else:
                device = "cpu"
        return torch.device(device)

    def eval(self) -> None:
        self.force_train = False

    def train(self) -> None:
        self.force_train = True

    def set_loss_y(self, y: NDNumber, **kwargs: Any) -> None:
        class_weight = self.class_weight
        if class_weight is not None:
            if self.class_weight == "balanced":
                y_w = y
                if y.ndim == 2:
                    y_w = y.argmax(1)
                class_weight = torch.Tensor(
                    1 - torch.unique(y_w, return_counts=True)[1] / len(y)
                )

            else:
                class_weight = torch.Tensor(self.class_weight)

        class_weight = class_weight.to(self.device)
        if self.objective == "regression":
            loss_func = nn.MSELoss(**kwargs)
        elif self.objective == "classification":
            loss_func = nn.CrossEntropyLoss(weight=class_weight, **kwargs)
        else:
            if hasattr(self, "num_classes") and (self.num_classes == 1):
                class_weight = torch.Tensor(class_weight[1] / class_weight[0])
            loss_func = nn.BCEWithLogitsLoss(pos_weight=class_weight, **kwargs)

        self.set_loss(loss_func)

    def fit(
        self,
        x: NDFloat,
        y: NDNumber,
        x_val: Optional[NDFloat] = None,
        y_val: Optional[NDNumber] = None,
        custom_train_dataloader=None,
        custom_val_dataloader=None,
        scaler=None
    ):
        # if reset_weight:
        #     self.reset_all_weights()

        optimizer = optim.AdamW(
            filter(lambda p: p.requires_grad, self.model.parameters()),
            lr=self.learning_rate,
            weight_decay=self.weight_decay,
        )

        if x_val is None:
            (
                _,
                count_idx,
                count,
            ) = np.unique(y, return_counts=True, return_index=True)
            if np.sum(count == 1) > 0:
                to_delete = count_idx[np.where(count == 1)[0]]
                x = np.delete(x, to_delete, axis=0)
                y = np.delete(y, to_delete, axis=0)

            x, x_val, y, y_val = train_test_split(
                x, y, test_size=0.2, random_state=42, stratify=y
            )

        if self.is_text:
            x = torch.tensor(x).int()
            x_val = torch.tensor(x_val).int()
        else:
            if isinstance(x, pd.DataFrame):
                x = x.values
            if isinstance(x_val, pd.DataFrame):
                x_val = x_val.values
            x = torch.tensor(x).float()
            x_val = torch.tensor(x_val).float()

        y = torch.tensor(y)
        y_val = torch.tensor(y_val)

        if (y.ndim == 1) or (y.shape[1] == 1):
            y_train = torch.column_stack((1 - y, y))
        if (y_val.ndim == 1) or (y_val.shape[1] == 1):
            y_val = torch.column_stack((1 - y_val, y_val))

        self.set_loss_y(y_train)
        loss_func = self.loss_func

        y_train = y_train.float()
        y_val = y_val.float()

        # train_loader = self.data_loader.get_dataloader(
        #     x, y, True, self.batch_size
        # )

        if custom_train_dataloader:
            if hasattr(custom_train_dataloader,"dataset"):
                custom_train_dataloader.dataset.tensors = x.to(self.device), y_train.to(self.device)
            else:
                custom_train_dataloader.tensors = x.to(self.device), y_train.to(self.device)
            train_loader = custom_train_dataloader
            train_loader.scaler = scaler
            train_loader.transform_y = True
        else:
            train_loader = FastTensorDataLoader(
                x.to(self.device),
                y_train.to(self.device),
                batch_size=self.batch_size,
                shuffle=True,
            )

        if custom_val_dataloader:
            custom_val_dataloader.dataset.tensors = x_val.to(self.device),
            y_val.to(self.device)
            val_loader = custom_val_dataloader
            train_loader.scaler = scaler
        else:
            val_loader = FastTensorDataLoader(
                x_val.to(self.device),
                y_val.to(self.device),
                batch_size=self.val_batch_size,
                shuffle=False,
            )

        min_val_loss = float("inf")
        min_val_loss_idx = 0

        loss_history = []
        val_loss_history = []

        for epoch in range(self.epochs):
            for i, (batch_x, batch_y) in enumerate(train_loader):

                out = self.model(batch_x.to(self.device))

                if (
                    self.objective == "regression"
                    or self.objective == "binary"
                ):
                    out = out.squeeze()

                loss = loss_func(out, batch_y.to(self.device))
                loss_history.append(loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                logger.debug(f"Batch {i}")

            # Early Stopping
            if self.early_stopping_rounds > 0:
                val_loss = 0.0
                val_dim = 0
                with torch.no_grad():
                    for val_i, (batch_val_x, batch_val_y) in enumerate(
                        val_loader
                    ):
                        out = self.model(batch_val_x.to(self.device))

                        if (
                            self.objective == "regression"
                            or self.objective == "binary"
                        ):
                            out = out.squeeze()

                        val_loss += loss_func(
                            out, batch_val_y.to(self.device)
                        ).item()
                        val_dim += 1

                val_loss /= val_dim
                val_loss_history.append(val_loss)

                # logger.debug("Epoch %d, Val Loss: %.5f" % (epoch, val_loss))
                print("Epoch %d, Val Loss: %.5f" % (epoch, val_loss))

                if val_loss < min_val_loss:
                    min_val_loss = val_loss
                    min_val_loss_idx = epoch

                    # Save the currently best model
                    self.save_best_weights()

                if min_val_loss_idx + self.early_stopping_rounds < epoch:
                    logger.debug(
                        "Validation loss has not improved for %d steps!"
                        % self.early_stopping_rounds
                    )
                    logger.debug("Early stopping applies.")
                    break

        self.model.__hash__()
        # Load best model
        if self.early_stopping_rounds > 0:
            self.load_best_weights()
        return loss_history, val_loss_history

    def predict(self, x):
        if self.objective == "regression":
            self.predictions = self.predict_helper(x)
        else:
            self.predict_proba(x)
            self.predictions = np.argmax(self.prediction_probabilities, axis=1)

        return self.predictions

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        probas = self.predict_helper(x)

        # If binary task returns only probability for the true class,
        # adapt it to return (N x 2)
        if probas.shape[1] == 1:
            probas = np.concatenate((1 - probas, probas), 1)

        self.prediction_probabilities = probas
        return self.prediction_probabilities

    def get_logits(self, x: torch.Tensor, with_grad: bool) -> torch.Tensor:
        if not with_grad:
            with torch.no_grad():
                # with_grad = True to avoid recursion
                return self.get_logits(x, with_grad=True)

        preds = self.model(x)
        if self.objective == "binary":
            preds = torch.sigmoid(preds)
        return preds

    def predict_helper(
        self,
        x: Union[NDFloat, torch.Tensor, pd.DataFrame],
        load_all_gpu: bool = False,
    ) -> NDFloat:
        if self.force_train:
            self.model.train()
        else:
            self.model.eval()

        if isinstance(x, pd.DataFrame):
            x = x.values

        if self.is_text:
            x = torch.tensor(x).int()
        else:
            x = torch.tensor(x).float()

        if load_all_gpu:
            x = x.to(self.device)

        # test_dataset = TensorDataset(x)
        # test_loader = DataLoader(
        #     dataset=test_dataset,
        #     batch_size=self.val_batch_size,
        #     shuffle=False,
        #     num_workers=0,
        # )

        test_loader = FastTensorDataLoader(
            x.to(self.device), batch_size=self.val_batch_size, shuffle=False
        )
        predictions = []
        with torch.no_grad():
            for batch_x in test_loader:
                if load_all_gpu:
                    preds = self.model(batch_x[0])
                else:
                    preds = self.model(batch_x[0].to(self.device))

                if self.objective == "binary":
                    preds = torch.sigmoid(preds)
                if self.device == "cuda":
                    torch.cuda.synchronize()
                predictions.append(preds.detach().cpu().numpy())
        return np.concatenate(predictions)

    def save(self, path: str) -> None:
        kwargs_save = deepcopy(self.constructor_kwargs)
        if "x_metadata" in kwargs_save:
            kwargs_save.pop("x_metadata")

        if "scaler" in kwargs_save:
            kwargs_save.pop("scaler")

        args_path = f"{path}/args.json"
        save_json(kwargs_save, parent_exists(args_path))

        weigths_path = f"{path}/weights.pt"
        torch.save(self.model.state_dict(), parent_exists(weigths_path))

    @classmethod
    def load_class(cls, path: str, **kwargs: Dict[str, Any]) -> BaseModelTorch:

        args_path = f"{path}/args.json"
        weigths_path = f"{path}/weights.pt"

        load_kwargs = load_json(args_path)
        args = {**load_kwargs, **kwargs}
        model = cls(**args)
        model.model.load_state_dict(torch.load(weigths_path))

        return model

    def save_best_weights(self) -> None:
        filename = parent_exists(f"./tmp/{self.uuid}_best.pt")
        torch.save(self.model.state_dict(), filename)

    def load_best_weights(self) -> None:
        filename = f"./tmp/{self.uuid}_best.pt"
        state_dict = torch.load(filename)
        self.model.load_state_dict(state_dict)

    def load(self, path: str) -> None:
        state_dict = torch.load(f"{path}/weights.pt")
        if self.scaler is None:
            model = self.model
        else:
            model = self.model[1]
        model.load_state_dict(state_dict)

    def get_model_size(self) -> int:
        model_size = sum(
            t.numel() for t in self.model.parameters() if t.requires_grad
        )
        return model_size

    def reset_all_weights(self) -> None:
        """
        refs:
            - https://discuss.pytorch.org/t/how-to-re-set-alll-parameters-in
            -a-network/20819/6
            - https://stackoverflow.com/questions/63627997/reset-parameters
            -of-a-neural-network-in-pytorch
            - https://pytorch.org/docs/stable/generated/torch.nn.Module.html
        """

        @torch.no_grad()
        def weight_reset(m: nn.Module) -> None:
            # - check if the current module has reset_parameters & if it's
            # callabed called it on m
            reset_parameters = getattr(m, "reset_parameters", None)
            if callable(reset_parameters):
                m.reset_parameters()

        # Applies fn recursively to every submodule see:
        # https://pytorch.org/docs/stable/generated/torch.nn.Module.html
        self.model.apply(fn=weight_reset)

    def set_dataloader(self, dataloader) -> None:
        self.data_loader = dataloader

    def set_loss(self, loss) -> None:
        self.loss_func = loss

    def set_experiment(self, experiment):
        self.experiment = experiment



# -*- coding:utf-8 -*-
"""
Author:
    Weichen Shen,weichenswc@163.com
"""

from collections import OrderedDict, namedtuple, defaultdict
from itertools import chain

import torch
import torch.nn as nn
import numpy as np

from .layers.sequence import SequencePoolingLayer
from .layers.utils import concat_fun

DEFAULT_GROUP_NAME = "default_group"


class SparseFeat(namedtuple('SparseFeat',
                            ['name', 'vocabulary_size', 'embedding_dim', 'use_hash', 'dtype', 'embedding_name',
                             'group_name'])):
    __slots__ = ()

    def __new__(cls, name, vocabulary_size, embedding_dim=4, use_hash=False, dtype="int32", embedding_name=None,
                group_name=DEFAULT_GROUP_NAME):
        if embedding_name is None:
            embedding_name = name
        if embedding_dim == "auto":
            embedding_dim = 6 * int(pow(vocabulary_size, 0.25))
        if use_hash:
            print(
                "Notice! Feature Hashing on the fly currently is not supported in torch version,you can use tensorflow version!")
        return super(SparseFeat, cls).__new__(cls, name, vocabulary_size, embedding_dim, use_hash, dtype,
                                              embedding_name, group_name)

    def __hash__(self):
        return self.name.__hash__()


class VarLenSparseFeat(namedtuple('VarLenSparseFeat',
                                  ['sparsefeat', 'maxlen', 'combiner', 'length_name'])):
    __slots__ = ()

    def __new__(cls, sparsefeat, maxlen, combiner="mean", length_name=None):
        return super(VarLenSparseFeat, cls).__new__(cls, sparsefeat, maxlen, combiner, length_name)

    @property
    def name(self):
        return self.sparsefeat.name

    @property
    def vocabulary_size(self):
        return self.sparsefeat.vocabulary_size

    @property
    def embedding_dim(self):
        return self.sparsefeat.embedding_dim

    @property
    def use_hash(self):
        return self.sparsefeat.use_hash

    @property
    def dtype(self):
        return self.sparsefeat.dtype

    @property
    def embedding_name(self):
        return self.sparsefeat.embedding_name

    @property
    def group_name(self):
        return self.sparsefeat.group_name

    def __hash__(self):
        return self.name.__hash__()


class DenseFeat(namedtuple('DenseFeat', ['name', 'dimension', 'dtype'])):
    __slots__ = ()

    def __new__(cls, name, dimension=1, dtype="float32"):
        return super(DenseFeat, cls).__new__(cls, name, dimension, dtype)

    def __hash__(self):
        return self.name.__hash__()


def get_feature_names(feature_columns):
    features = build_input_features(feature_columns)
    return list(features.keys())


# def get_inputs_list(inputs):
#     return list(chain(*list(map(lambda x: x.values(), filter(lambda x: x is not None, inputs)))))


def build_input_features(feature_columns):
    # Return OrderedDict: {feature_name:(start, start+dimension)}

    features = OrderedDict()

    start = 0
    for feat in feature_columns:
        feat_name = feat.name
        if feat_name in features:
            continue
        if isinstance(feat, SparseFeat):
            features[feat_name] = (start, start + 1)
            start += 1
        elif isinstance(feat, DenseFeat):
            features[feat_name] = (start, start + feat.dimension)
            start += feat.dimension
        elif isinstance(feat, VarLenSparseFeat):
            features[feat_name] = (start, start + feat.maxlen)
            start += feat.maxlen
            if feat.length_name is not None and feat.length_name not in features:
                features[feat.length_name] = (start, start + 1)
                start += 1
        else:
            raise TypeError("Invalid feature column type,got", type(feat))
    return features


def combined_dnn_input(sparse_embedding_list, dense_value_list):
    if len(sparse_embedding_list) > 0 and len(dense_value_list) > 0:
        sparse_dnn_input = torch.flatten(
            torch.cat(sparse_embedding_list, dim=-1), start_dim=1)
        dense_dnn_input = torch.flatten(
            torch.cat(dense_value_list, dim=-1), start_dim=1)
        return concat_fun([sparse_dnn_input, dense_dnn_input])
    elif len(sparse_embedding_list) > 0:
        return torch.flatten(torch.cat(sparse_embedding_list, dim=-1), start_dim=1)
    elif len(dense_value_list) > 0:
        return torch.flatten(torch.cat(dense_value_list, dim=-1), start_dim=1)
    else:
        raise NotImplementedError


def get_varlen_pooling_list(embedding_dict, features, feature_index, varlen_sparse_feature_columns, device):
    varlen_sparse_embedding_list = []
    for feat in varlen_sparse_feature_columns:
        seq_emb = embedding_dict[feat.name]
        if feat.length_name is None:
            seq_mask = features[:, feature_index[feat.name][0]:feature_index[feat.name][1]].long() != 0

            emb = SequencePoolingLayer(mode=feat.combiner, supports_masking=True, device=device)(
                [seq_emb, seq_mask])
        else:
            seq_length = features[:, feature_index[feat.length_name][0]:feature_index[feat.length_name][1]].long()
            emb = SequencePoolingLayer(mode=feat.combiner, supports_masking=False, device=device)(
                [seq_emb, seq_length])
        varlen_sparse_embedding_list.append(emb)
    return varlen_sparse_embedding_list


def create_embedding_matrix(feature_columns, init_std=0.0001, linear=False, sparse=False, device='cpu'):
    # Return nn.ModuleDict: for sparse features, {embedding_name: nn.Embedding}
    # for varlen sparse features, {embedding_name: nn.EmbeddingBag}
    sparse_feature_columns = list(
        filter(lambda x: isinstance(x, SparseFeat), feature_columns)) if len(feature_columns) else []

    varlen_sparse_feature_columns = list(
        filter(lambda x: isinstance(x, VarLenSparseFeat), feature_columns)) if len(feature_columns) else []

    embedding_dict = nn.ModuleDict(
        {feat.embedding_name: nn.Embedding(feat.vocabulary_size, feat.embedding_dim if not linear else 1, sparse=sparse)
         for feat in
         sparse_feature_columns + varlen_sparse_feature_columns}
    )

    # for feat in varlen_sparse_feature_columns:
    #     embedding_dict[feat.embedding_name] = nn.EmbeddingBag(
    #         feat.dimension, embedding_size, sparse=sparse, mode=feat.combiner)

    for tensor in embedding_dict.values():
        nn.init.normal_(tensor.weight, mean=0, std=init_std)

    return embedding_dict.to(device)


def embedding_lookup(X, sparse_embedding_dict, sparse_input_dict, sparse_feature_columns, return_feat_list=(),
                     mask_feat_list=(), to_list=False):
    """
        Args:
            X: input Tensor [batch_size x hidden_dim]
            sparse_embedding_dict: nn.ModuleDict, {embedding_name: nn.Embedding}
            sparse_input_dict: OrderedDict, {feature_name:(start, start+dimension)}
            sparse_feature_columns: list, sparse features
            return_feat_list: list, names of feature to be returned, defualt () -> return all features
            mask_feat_list, list, names of feature to be masked in hash transform
        Return:
            group_embedding_dict: defaultdict(list)
    """
    group_embedding_dict = defaultdict(list)
    for fc in sparse_feature_columns:
        feature_name = fc.name
        embedding_name = fc.embedding_name
        if (len(return_feat_list) == 0 or feature_name in return_feat_list):
            # TODO: add hash function
            # if fc.use_hash:
            #     raise NotImplementedError("hash function is not implemented in this version!")
            lookup_idx = np.array(sparse_input_dict[feature_name])
            input_tensor = X[:, lookup_idx[0]:lookup_idx[1]].long()
            emb = sparse_embedding_dict[embedding_name](input_tensor)
            group_embedding_dict[fc.group_name].append(emb)
    if to_list:
        return list(chain.from_iterable(group_embedding_dict.values()))
    return group_embedding_dict


def varlen_embedding_lookup(X, embedding_dict, sequence_input_dict, varlen_sparse_feature_columns):
    varlen_embedding_vec_dict = {}
    for fc in varlen_sparse_feature_columns:
        feature_name = fc.name
        embedding_name = fc.embedding_name
        if fc.use_hash:
            # lookup_idx = Hash(fc.vocabulary_size, mask_zero=True)(sequence_input_dict[feature_name])
            # TODO: add hash function
            lookup_idx = sequence_input_dict[feature_name]
        else:
            lookup_idx = sequence_input_dict[feature_name]
        varlen_embedding_vec_dict[feature_name] = embedding_dict[embedding_name](
            X[:, lookup_idx[0]:lookup_idx[1]].long())  # (lookup_idx)

    return varlen_embedding_vec_dict


def get_dense_input(X, features, feature_columns):
    dense_feature_columns = list(filter(lambda x: isinstance(
        x, DenseFeat), feature_columns)) if feature_columns else []
    dense_input_list = []
    for fc in dense_feature_columns:
        lookup_idx = np.array(features[fc.name])
        input_tensor = X[:, lookup_idx[0]:lookup_idx[1]].float()
        dense_input_list.append(input_tensor)
    return dense_input_list


def maxlen_lookup(X, sparse_input_dict, maxlen_column):
    if maxlen_column is None or len(maxlen_column)==0:
        raise ValueError('please add max length column for VarLenSparseFeat of DIN/DIEN input')
    lookup_idx = np.array(sparse_input_dict[maxlen_column[0]])
    return X[:, lookup_idx[0]:lookup_idx[1]].long()



from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from mlc.models.model import Model
from mlc.transformers.transformer import Transformer


class Pipeline(Model):
    def __init__(self, steps: List[Union[Transformer, Model]]):
        super().__init__(name=steps[-1].name, objective=steps[-1].objective)
        self.steps = steps
        # For compatibility

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: Union[None, np.ndarray] = None,
        y_val: Union[None, np.ndarray] = None,
    ):
        for step in self.steps[:-1]:
            x = step.fit_transform(x, y)
        self.steps[-1].fit(x, y, x_val, y_val)
        return self

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        x = self.transform(x)
        return self.steps[-1].predict_proba(x)

    def predict(
        self, x: Union[pd.DataFrame, np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray]:
        for step in self.steps[:-1]:
            x = step.transform(x)
        return self.steps[-1].predict(x)

    def load(self, path):
        for i, step in enumerate(self.steps):
            step.load(f"{path}/{i}_{step.name}.model")

    def save(self, path):
        Path(path).mkdir(parents=True, exist_ok=True)
        for i, step in enumerate(self.steps):
            step.save(f"{path}/{i}_{step.name}.model")

    def clone(self):
        return self.__class__([step.clone() for step in self.steps])

    def __getitem__(self, item):
        return self.steps[item]

    def transform(self, x):
        for step in self.steps[:-1]:
            x = step.transform(x)
        return x



from abc import abstractmethod
from typing import Any, List, Union


class Backend:

    # Values

    @abstractmethod
    def get_zeros(self, operands: Any) -> Any:
        pass

    @abstractmethod
    def constant(self, value: Union[int, float]) -> Any:
        pass

    @abstractmethod
    def feature(self, x: Any, feature_id: int) -> Any:
        pass

    # Math operations
    @abstractmethod
    def math_operation(
        self, operator: str, left_operand: Any, right_operand: Any
    ) -> Any:
        pass

    @abstractmethod
    def safe_division(
        self, dividend: Any, divisor: Any, safe_value: Any
    ) -> Any:
        pass

    @abstractmethod
    def log(self, operand: Any, safe_value: Any = None) -> Any:
        pass

    @abstractmethod
    def many_sum(self, operands: List[Any]) -> Any:
        pass

    # Comparisons
    @abstractmethod
    def less_equal_constraint(
        self, left_operand: Any, right_operand: Any
    ) -> Any:
        pass

    @abstractmethod
    def equal_constraint(self, left_operand: Any, right_operand: Any) -> Any:
        pass

    @abstractmethod
    def less_constraint(self, left_operand: Any, right_operand: Any) -> Any:
        pass

    @abstractmethod
    def or_constraint(self, operands: List[Any]) -> Any:
        pass

    @abstractmethod
    def and_constraint(self, operands: List[Any]) -> Any:
        pass

    # Extension

    @abstractmethod
    def count(self, operands: List[Any], inverse: bool) -> Any:
        pass



import torch
import numpy as np


def embed_data_mask(x_categ, x_cont, cat_mask, con_mask, model, vision_dset=False):
    device = x_cont.device
    x_categ = x_categ + model.categories_offset.type_as(x_categ)
    x_categ_enc = model.embeds(x_categ)
    n1, n2 = x_cont.shape
    _, n3 = x_categ.shape
    if model.cont_embeddings == 'MLP':
        x_cont_enc = torch.empty(n1, n2, model.dim)
        for i in range(model.num_continuous):
            x_cont_enc[:, i, :] = model.simple_MLP[i](x_cont[:, i])
    else:
        raise Exception('This case should not work!')

    x_cont_enc = x_cont_enc.to(device)
    cat_mask_temp = cat_mask + model.cat_mask_offset.type_as(cat_mask)
    con_mask_temp = con_mask + model.con_mask_offset.type_as(con_mask)

    cat_mask_temp = model.mask_embeds_cat(cat_mask_temp)
    con_mask_temp = model.mask_embeds_cont(con_mask_temp)
    x_categ_enc[cat_mask == 0] = cat_mask_temp[cat_mask == 0]
    x_cont_enc[con_mask == 0] = con_mask_temp[con_mask == 0]

    if vision_dset:
        pos = np.tile(np.arange(x_categ.shape[-1]), (x_categ.shape[0], 1))
        pos = torch.from_numpy(pos).to(device)
        pos_enc = model.pos_encodings(pos)
        x_categ_enc += pos_enc

    return x_categ, x_categ_enc, x_cont_enc


def mixup_data(x1, x2, lam=1.0, y=None, use_cuda=True):
    '''Returns mixed inputs, pairs of targets'''

    batch_size = x1.size()[0]
    if use_cuda:
        index = torch.randperm(batch_size).cuda()
    else:
        index = torch.randperm(batch_size)

    mixed_x1 = lam * x1 + (1 - lam) * x1[index, :]
    mixed_x2 = lam * x2 + (1 - lam) * x2[index, :]
    if y is not None:
        y_a, y_b = y, y[index]
        return mixed_x1, mixed_x2, y_a, y_b

    return mixed_x1, mixed_x2


def add_noise(x_categ, x_cont, noise_params={'noise_type': ['cutmix'], 'lambda': 0.1}):
    lam = noise_params['lambda']
    device = x_categ.device
    batch_size = x_categ.size()[0]

    if 'cutmix' in noise_params['noise_type']:
        index = torch.randperm(batch_size)
        cat_corr = torch.from_numpy(np.random.choice(2, (x_categ.shape), p=[lam, 1 - lam])).to(device)
        con_corr = torch.from_numpy(np.random.choice(2, (x_cont.shape), p=[lam, 1 - lam])).to(device)
        x1, x2 = x_categ[index, :], x_cont[index, :]
        x_categ_corr, x_cont_corr = x_categ.clone().detach(), x_cont.clone().detach()
        x_categ_corr[cat_corr == 0] = x1[cat_corr == 0]
        x_cont_corr[con_corr == 0] = x2[con_corr == 0]
        return x_categ_corr, x_cont_corr
    elif noise_params['noise_type'] == 'missing':
        x_categ_mask = np.random.choice(2, (x_categ.shape), p=[lam, 1 - lam])
        x_cont_mask = np.random.choice(2, (x_cont.shape), p=[lam, 1 - lam])
        x_categ_mask = torch.from_numpy(x_categ_mask).to(device)
        x_cont_mask = torch.from_numpy(x_cont_mask).to(device)
        return torch.mul(x_categ, x_categ_mask), torch.mul(x_cont, x_cont_mask)

    else:
        print("yet to write this")



from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import optuna
import pandas as pd
import torch
import torch.nn as nn
from mlc.dataloaders.fast_dataloader import FastTensorDataLoader

from mlc.models.model import Model
from mlc.models.torch_models import BaseModelTorch
from mlc.transformers.tab_scaler import TabScaler
from mlc.utils import to_torch_number

from .deepfm_lib.inputs import DenseFeat, SparseFeat
from .deepfm_lib.models.deepfm import DeepFM as DeepFMModel

"""
    DeepFM: A Factorization-Machine based Neural Network for CTR Prediction
     (https://www.ijcai.org/proceedings/2017/0239.pdf)

    Code adapted from: https://github.com/shenweichen/DeepCTR-Torch
"""


class BalancedBCELoss(torch.nn.BCEWithLogitsLoss):
    def __init__(self, y: torch.Tensor, **args: Any):
        self.weights = None
        y = np.array(y)
        y_class, y_occ = np.unique(y, return_counts=True)
        self.weights = dict(zip(y_class, y_occ / len(y)))
        super(BalancedBCELoss, self).__init__(**args)

    def forward(
        self, input: torch.Tensor, target: torch.Tensor, reduction: Any = None
    ) -> torch.Tensor:

        if self.weights is None:
            return super(BalancedBCELoss, self).forward(input, target)

        negative_inputs_mask = target == 0
        positive_inputs_mask = target == 1

        positive_inputs, positive_targets = (
            input[positive_inputs_mask],
            input[positive_inputs_mask],
        )
        positive_loss = super(BalancedBCELoss, self).forward(
            positive_inputs, positive_targets
        )
        negative_inputs, negative_targets = (
            input[negative_inputs_mask],
            input[negative_inputs_mask],
        )
        negative_loss = super(BalancedBCELoss, self).forward(
            negative_inputs, negative_targets
        )

        return positive_loss / self.weights.get(
            1
        ) + negative_loss / self.weights.get(0)


def get_fixlen_feature_columns(
    cat_idx: List[int],
    cat_dims: List[int],
    num_features: int,
    num_dense_features: int = None,
) -> List[Union[DenseFeat, SparseFeat]]:
    if cat_idx:
        if (
            num_dense_features
            and num_dense_features > 0
            and num_dense_features + len(cat_idx) < num_features
        ):
            # i.e. we have one hot encoded the features, using binary sparse
            dense_features = list(
                range(
                    num_features - num_dense_features,
                    num_features,
                )
            )
            binary_featues = list(range(0, num_dense_features))
            fixlen_feature_columns = [
                SparseFeat(str(feat), 2) for feat in binary_featues
            ] + [
                DenseFeat(
                    str(feat),
                    1,
                )
                for feat in dense_features
            ]
        else:
            dense_features = list(set(range(num_features)) - set(cat_idx))
            fixlen_feature_columns = [
                SparseFeat(str(feat), cat_dims[idx])
                for idx, feat in enumerate(cat_idx)
            ] + [
                DenseFeat(
                    str(feat),
                    1,
                )
                for feat in dense_features
            ]

    else:
        # Add dummy sparse feature, otherwise it will crash...
        fixlen_feature_columns = [SparseFeat("dummy", 1)] + [
            DenseFeat(
                str(feat),
                1,
            )
            for feat in range(num_features)
        ]

    return fixlen_feature_columns


class DeepFMWrapper(nn.Module):
    def __init__(self, num_features, cat_idx, feature_index) -> None:

        self.num_features = num_features
        self.cat_idx = cat_idx
        self.feature_index = feature_index
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        X_dict = {str(name): x[:, name] for name in range(self.num_features)}

        # Adding dummy spare feature
        if not self.cat_idx:
            X_dict["dummy"] = torch.zeros(x.shape[0])

        x_stacked = [X_dict[feature] for feature in self.feature_index]
        for i in range(len(x_stacked)):
            if len(x_stacked[i].shape) == 1:
                x_stacked[i] = x_stacked[i].unsqueeze(-1).to(x.device)

        x = torch.column_stack(x_stacked)
        return x


class DeepFMOutputLayer(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.column_stack([1 - x, x])


class DeepFM(BaseModelTorch):
    def __init__(
        self,
        objective: str,
        x_metadata: pd.DataFrame,
        batch_size: int,
        epochs: int,
        early_stopping_rounds: int,
        dnn_dropout: float,
        name: str = "deepfm",
        num_dense_features: Optional[int] = None,
        num_classes: int = 2,
        scaler: Optional[TabScaler] = None,
        **kwargs,
    ):
        # Parameters
        self.objective = objective
        self.x_metadata = x_metadata
        self.batch_size = batch_size
        self.epochs = epochs
        self.early_stopping_rounds = early_stopping_rounds
        self.dnn_dropout = dnn_dropout
        self.num_dense_features = num_dense_features
        self.num_classes = num_classes

        if scaler is not None:
            self.scaler = TabScaler(num_scaler="min_max", one_hot_encode=False)
            self.scaler.fit_scaler_data(scaler.get_scaler_data())

        super().__init__(
            objective=objective,
            x_metadata=x_metadata,
            batch_size=batch_size,
            epochs=epochs,
            early_stopping_rounds=early_stopping_rounds,
            dnn_dropout=dnn_dropout,
            name=name,
            num_dense_features=num_dense_features,
            num_classes=num_classes,
            scaler= scaler,
            **kwargs,
        )

        # Generated

        # Parameters check
        self.two_d_output = False
        if self.objective == "classification":
            if self.num_classes != 2:
                raise ValueError(
                    "DeepFM not yet implemented for classification"
                )
            else:
                self.objective = "binary"
                self.num_classes = 1
                self.two_d_output = True

        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_features = x_metadata.shape[0]
        self.cat_idx = np.where(x_metadata["type"] == "cat")[0].tolist()
        self.num_idx = list(set(range(self.num_features)) - set(self.cat_idx))
        self.cat_dims = [
            (int(x_metadata.iloc[i]["max"]) + 1) for i in self.cat_idx
        ]

        # self.dataset = args.dataset

        fixlen_feature_columns = get_fixlen_feature_columns(
            self.cat_idx,
            self.cat_dims,
            self.num_features,
            num_dense_features=self.num_dense_features,
        )

        self.model: nn.Module = DeepFMModel(
            linear_feature_columns=fixlen_feature_columns,
            dnn_feature_columns=fixlen_feature_columns,
            task=self.objective,
            device=self.device,
            dnn_dropout=self.dnn_dropout,
            gpus=self.gpus,
        )

        self.wrapper_model = nn.Sequential(
            DeepFMWrapper(
                self.num_features, self.cat_idx, self.model.feature_index
            ),
            self.model,
            DeepFMOutputLayer(),
        )

        if scaler is not None:
            self.wrapper_model = nn.Sequential(
                self.scaler.get_transorm_nn(), self.wrapper_model
            )

        # For compatibility
        self.experiment = None

    def fit(self, X, y, X_val=None, y_val=None,
        custom_train_dataloader=None,
        custom_val_dataloader=None):

        if self.scaler is not None:
            X = self.scaler.transform(X)
            X_val = self.scaler.transform(X_val)

        X = np.array(X, dtype=float)
        X_dict = {str(name): X[:, name] for name in range(self.num_features)}

        X_val = np.array(X_val, dtype=float)
        X_val_dict = {
            str(name): X_val[:, name] for name in range(self.num_features)
        }

        if self.objective == "binary":
            self.set_loss_y(to_torch_number(y), reduction="sum")
            loss = lambda y, y2, reduction: self.loss_func(y, y2)
            metric = "binary_crossentropy"
            labels = [0, 1]
        elif self.objective == "regression":
            loss = "mse"
            metric = "mse"
            labels = None

        self.model.compile(
            optimizer=torch.optim.AdamW(self.model.parameters()),
            loss=loss,
            metrics=[metric],
        )

        # Adding dummy spare feature
        if not self.cat_idx:
            X_dict["dummy"] = np.zeros(X.shape[0])
            X_val_dict["dummy"] = np.zeros(X_val.shape[0])

        loss_history, val_loss_history = self.model.fit(
            X_dict,
            y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            validation_data=(X_val_dict, y_val),
            labels=labels,
            early_stopping=True,
            patience=self.early_stopping_rounds,
            custom_train_dataloader = custom_train_dataloader,
            custom_val_dataloader=custom_val_dataloader,
            scaler = self.scaler

        )

        if self.experiment is not None:
            for lo in loss_history:
                self.experiment.log_metric("train_loss", lo)

            for lo in val_loss_history:
                self.experiment.log_metric("validation_loss", lo)

        return loss_history, val_loss_history

    def get_logits(self, x: torch.Tensor, with_grad: bool) -> torch.Tensor:
        self.model.eval()
        if not with_grad:
            with torch.no_grad():
                # with_grad = True to avoid recursion
                return self.get_logits(x, with_grad=True)

        # Warning: use wrapper model
        preds = self.wrapper_model(x.to(self.device))
        return preds

    def predict_helper(self, x, keep_grad=False):
        self.model.eval()

        x = np.array(x, dtype=float)
        x = torch.tensor(x).float().to(self.device)
        test_loader = FastTensorDataLoader(x, batch_size=self.val_batch_size, shuffle=False)

        predictions = []
        for batch_x in test_loader:
            preds = self.get_logits(batch_x[0], with_grad=keep_grad)
            predictions.append(preds.cpu())

        return np.concatenate(predictions)

    @staticmethod
    def define_trial_parameters(
        trial: optuna.Trial, trial_params: Dict[str, Any]
    ) -> Dict[str, Any]:
        params = {
            "dnn_dropout": trial.suggest_float("dnn_dropout", 0, 0.9),
        }
        return params

        # dnn_dropout, l2_reg_linear, l2_reg_embedding,
        # l2_reg_dnn, dnn_hidden_units?


models: List[Tuple[str, Type[Model]]] = [("deepfm", DeepFM)]
